# RCAIDE/Library/Methods/Geodesics/Geodesics.py
# 
# 
# Adapted:  Oct 2024, A. Molloy
#
# This is a built-in adaptation of the Geopy distance claculation capabilities. It primarily serves to calcualte distances
# lat/long coordinate points. Minor adjustments have been made to better fit within the RCAIDE frameowkr (for instance, results
# only returned in kilometers).
#
# Distinct classes that are included in one palce here include the original Constants, Distance, Geodesic, 
# GeodesicCapability, Geodesic_Calculate, and Math classes.

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
import math
import sys

"""Evaluate the thrust produced by the energy network.

Assumptions:
    WGS-84 Ellipsoid model 

Source:
    Karney, C. F., (2022) Geopy Python Code [source code]. https://geographiclib.sourceforge.io/

Args:
    
Returns:
     
""" 

# geodesic.py
#
# This is a rather literal translation of the GeographicLib::Geodesic class to
# python.  See the documentation for the C++ class for more information at
#
#    https://geographiclib.sourceforge.io/html/annotated.html
#
# The algorithms are derived in
#
#    Charles F. F. Karney,
#    Algorithms for geodesics, J. Geodesy 87, 43-55 (2013),
#    https://doi.org/10.1007/s00190-012-0578-z
#    Addenda: https://geographiclib.sourceforge.io/geod-addenda.html
#
# Copyright (c) Charles Karney (2011-2022) <charles@karney.com> and licensed
# under the MIT/X11 License.  For more information, see
# https://geographiclib.sourceforge.io/
######################################################################

class Math:
    """
    Additional math routines for GeographicLib.
    """

    @staticmethod
    def sq(x):
        """Square a number"""

        return x * x

    @staticmethod
    def cbrt(x):
        """Real cube root of a number"""

        return math.copysign(math.pow(abs(x), 1/3.0), x)

    @staticmethod
    def norm(x, y):
        """Private: Normalize a two-vector."""

        r = (math.sqrt(Math.sq(x) + Math.sq(y))
         # hypot is inaccurate for 3.[89].  Problem reported by agdhruv
         # https://github.com/geopy/geopy/issues/466 ; see
         # https://bugs.python.org/issue43088
         # Visual Studio 2015 32-bit has a similar problem.
         if (3, 8) <= sys.version_info < (3, 10)
         else math.hypot(x, y))
        return x/r, y/r

    @staticmethod
    def sum(u, v):
        """Error free transformation of a sum."""

        # Error free transformation of a sum.  Note that t can be the same as one
        # of the first two arguments.
        s = u + v
        up = s - v
        vpp = s - up
        up -= u
        vpp -= v
        t = s if s == 0 else 0.0 - (up + vpp)
        # u + v =       s      + t
        #       = round(u + v) + t
        return s, t

    @staticmethod
    def polyval(N, p, s, x):
        """Evaluate a polynomial."""

        y = float(0 if N < 0 else p[s]) # make sure the returned value is a float
        while N > 0:
            N -= 1; s += 1
            y = y * x + p[s]
        return y

    @staticmethod
    def AngRound(x):
        """Private: Round an angle so that small values underflow to zero."""

        # The makes the smallest gap in x = 1/16 - nextafter(1/16, 0) = 1/2^57
        # for reals = 0.7 pm on the earth if x is an angle in degrees.  (This
        # is about 1000 times more resolution than we get with angles around 90
        # degrees.)  We use this to avoid having to deal with near singular
        # cases when x is non-zero but tiny (e.g., 1.0e-200).
        z = 1/16.0
        y = abs(x)
        # The compiler mustn't "simplify" z - (z - y) to y
        if y < z: y = z - (z - y)
        return math.copysign(y, x)

    @staticmethod
    def remainder(x, y):
        """remainder of x/y in the range [-y/2, y/2]."""

        return math.remainder(x, y) if math.isfinite(x) else math.nan

    @staticmethod
    def AngNormalize(x):
        """reduce angle to [-180,180]"""

        y = Math.remainder(x, 360)
        return math.copysign(180.0, x) if abs(y) == 180 else y

    @staticmethod
    def LatFix(x):
        """replace angles outside [-90,90] by NaN"""

        return math.nan if abs(x) > 90 else x

    @staticmethod
    def AngDiff(x, y):
        """compute y - x and reduce to [-180,180] accurately"""

        d, t = Math.sum(Math.remainder(-x, 360), Math.remainder(y, 360))
        d, t = Math.sum(Math.remainder(d, 360), t)
        if d == 0 or abs(d) == 180:
            d = math.copysign(d, y - x if t == 0 else -t)
        return d, t

    @staticmethod
    def sincosd(x):
        """Compute sine and cosine of x in degrees."""

        r = math.fmod(x, 360) if math.isfinite(x) else math.nan
        q = 0 if math.isnan(r) else int(round(r / 90))
        r -= 90 * q; r = math.radians(r)
        s = math.sin(r); c = math.cos(r)
        q = q % 4
        if   q == 1: s, c =  c, -s
        elif q == 2: s, c = -s, -c
        elif q == 3: s, c = -c,  s
        c = c + 0.0
        if s == 0: s = math.copysign(s, x)
        return s, c

    @staticmethod
    def sincosde(x, t):
        """Compute sine and cosine of (x + t) in degrees with x in [-180, 180]"""

        q = int(round(x / 90)) if math.isfinite(x) else 0
        r = x - 90 * q; r = math.radians(Math.AngRound(r + t))
        s = math.sin(r); c = math.cos(r)
        q = q % 4
        if   q == 1: s, c =  c, -s
        elif q == 2: s, c = -s, -c
        elif q == 3: s, c = -c,  s
        c = c + 0.0
        if s == 0: s = math.copysign(s, x)
        return s, c
 
class GeodesicCapability:
    """
    Capability constants shared between Geodesic and GeodesicLine.
    """

    CAP_NONE = 0
    CAP_C1   = 1 << 0
    CAP_C1p  = 1 << 1
    CAP_C2   = 1 << 2
    CAP_C3   = 1 << 3
    CAP_C4   = 1 << 4
    CAP_ALL  = 0x1F
    CAP_MASK = CAP_ALL
    OUT_ALL  = 0x7F80
    OUT_MASK = 0xFF80             # Includes LONG_UNROLL
    EMPTY         = 0
    LATITUDE      = 1 << 7  | CAP_NONE
    LONGITUDE     = 1 << 8  | CAP_C3
    AZIMUTH       = 1 << 9  | CAP_NONE
    DISTANCE      = 1 << 10 | CAP_C1
    STANDARD      = LATITUDE | LONGITUDE | AZIMUTH | DISTANCE
    DISTANCE_IN   = 1 << 11 | CAP_C1 | CAP_C1p
    REDUCEDLENGTH = 1 << 12 | CAP_C1 | CAP_C2
    GEODESICSCALE = 1 << 13 | CAP_C1 | CAP_C2
    AREA          = 1 << 14 | CAP_C4
    LONG_UNROLL   = 1 << 15
    ALL           = OUT_ALL | CAP_ALL # Does not include LONG_UNROLL


class Constants:
    """
    Constants describing the WGS84 ellipsoid
    """
    WGS84_a = 6378137.0           # meters
    """the equatorial radius in meters of the WGS84 ellipsoid in meters"""
    WGS84_f = 1/298.257223563
    """the flattening of the WGS84 ellipsoid, 1/298.257223563"""

class Geodesic:
    """Solve geodesic problems"""

    GEOGRAPHICLIB_GEODESIC_ORDER = 6
    nA1_ = GEOGRAPHICLIB_GEODESIC_ORDER
    nC1_ = GEOGRAPHICLIB_GEODESIC_ORDER
    nC1p_ = GEOGRAPHICLIB_GEODESIC_ORDER
    nA2_ = GEOGRAPHICLIB_GEODESIC_ORDER
    nC2_ = GEOGRAPHICLIB_GEODESIC_ORDER
    nA3_ = GEOGRAPHICLIB_GEODESIC_ORDER
    nA3x_ = nA3_
    nC3_ = GEOGRAPHICLIB_GEODESIC_ORDER
    nC3x_ = (nC3_ * (nC3_ - 1)) // 2
    nC4_ = GEOGRAPHICLIB_GEODESIC_ORDER
    nC4x_ = (nC4_ * (nC4_ + 1)) // 2
    maxit1_ = 20
    maxit2_ = maxit1_ + sys.float_info.mant_dig + 10

    tiny_ = math.sqrt(sys.float_info.min)
    tol0_ = sys.float_info.epsilon
    tol1_ = 200 * tol0_
    tol2_ = math.sqrt(tol0_)
    tolb_ = tol0_ * tol2_
    xthresh_ = 1000 * tol2_

    CAP_NONE = GeodesicCapability.CAP_NONE
    CAP_C1   = GeodesicCapability.CAP_C1
    CAP_C1p  = GeodesicCapability.CAP_C1p
    CAP_C2   = GeodesicCapability.CAP_C2
    CAP_C3   = GeodesicCapability.CAP_C3
    CAP_C4   = GeodesicCapability.CAP_C4
    CAP_ALL  = GeodesicCapability.CAP_ALL
    CAP_MASK = GeodesicCapability.CAP_MASK
    OUT_ALL  = GeodesicCapability.OUT_ALL
    OUT_MASK = GeodesicCapability.OUT_MASK

    @staticmethod
    def _SinCosSeries(sinp, sinx, cosx, c):
        """Private: Evaluate a trig series using Clenshaw summation."""
        # Evaluate
        # y = sinp ? sum(c[i] * sin( 2*i    * x), i, 1, n) :
        #            sum(c[i] * cos((2*i+1) * x), i, 0, n-1)
        # using Clenshaw summation.  N.B. c[0] is unused for sin series
        # Approx operation count = (n + 5) mult and (2 * n + 2) add
        k = len(c)                  # Point to one beyond last element
        n = k - sinp
        ar = 2 * (cosx - sinx) * (cosx + sinx) # 2 * cos(2 * x)
        y1 = 0                                 # accumulators for sum
        if n & 1:
            k -= 1; y0 = c[k]
        else:
            y0 = 0
        # Now n is even
        n = n // 2
        while n:                    # while n--:
            n -= 1
            # Unroll loop x 2, so accumulators return to their original role
            k -= 1; y1 = ar * y0 - y1 + c[k]
            k -= 1; y0 = ar * y1 - y0 + c[k]
        return ( 2 * sinx * cosx * y0 if sinp # sin(2 * x) * y0
             else cosx * (y0 - y1) )      # cos(x) * (y0 - y1)
 

    @staticmethod
    def _A1m1f(eps):
        """Private: return A1-1."""
        coeff = [
        1, 4, 64, 0, 256,
    ]
        m = Geodesic.nA1_//2
        t = Math.polyval(m, coeff, 0, Math.sq(eps)) / coeff[m + 1]
        return (t + eps) / (1 - eps)

    @staticmethod
    def _C1f(eps, c):
        """Private: return C1."""
        coeff = [
        -1, 6, -16, 32,
      -9, 64, -128, 2048,
      9, -16, 768,
      3, -5, 512,
      -7, 1280,
      -7, 2048,
    ]
        eps2 = Math.sq(eps)
        d = eps
        o = 0
        for l in range(1, Geodesic.nC1_ + 1): # l is index of C1p[l]
            m = (Geodesic.nC1_ - l) // 2        # order of polynomial in eps^2
            c[l] = d * Math.polyval(m, coeff, o, eps2) / coeff[o + m + 1]
            o += m + 2
            d *= eps
            
    @staticmethod
    def _A2m1f(eps):
        """Private: return A2-1"""
        coeff = [
        -11, -28, -192, 0, 256,
    ]
        m = Geodesic.nA2_//2
        t = Math.polyval(m, coeff, 0, Math.sq(eps)) / coeff[m + 1]
        return (t - eps) / (1 + eps)

    @staticmethod
    def _C2f(eps, c):
        """Private: return C2"""
        coeff = [
        1, 2, 16, 32,
      35, 64, 384, 2048,
      15, 80, 768,
      7, 35, 512,
      63, 1280,
      77, 2048,
    ]
        eps2 = Math.sq(eps)
        d = eps
        o = 0
        for l in range(1, Geodesic.nC2_ + 1): # l is index of C2[l]
            m = (Geodesic.nC2_ - l) // 2        # order of polynomial in eps^2
            c[l] = d * Math.polyval(m, coeff, o, eps2) / coeff[o + m + 1]
            o += m + 2
            d *= eps

    def __init__(self, a, f):
        """Construct a Geodesic object

        :param a: the equatorial radius of the ellipsoid in meters
        :param f: the flattening of the ellipsoid

        An exception is thrown if *a* or the polar semi-axis *b* = *a* (1 -
        *f*) is not a finite positive quantity.

        """

        self.a = float(a)
        """The equatorial radius in meters (readonly)"""
        self.f = float(f)
        """The flattening (readonly)"""
        self._f1 = 1 - self.f
        self._e2 = self.f * (2 - self.f)
        self._ep2 = self._e2 / Math.sq(self._f1) # e2 / (1 - e2)
        self._n = self.f / ( 2 - self.f)
        self._b = self.a * self._f1
        # authalic radius squared
        self._c2 = (Math.sq(self.a) + Math.sq(self._b) *
                (1 if self._e2 == 0 else
                 (math.atanh(math.sqrt(self._e2)) if self._e2 > 0 else
                  math.atan(math.sqrt(-self._e2))) /
                 math.sqrt(abs(self._e2))))/2
        # The sig12 threshold for "really short".  Using the auxiliary sphere
        # solution with dnm computed at (bet1 + bet2) / 2, the relative error in
        # the azimuth consistency check is sig12^2 * abs(f) * min(1, 1-f/2) / 2.
        # (Error measured for 1/100 < b/a < 100 and abs(f) >= 1/1000.  For a given
        # f and sig12, the max error occurs for lines near the pole.  If the old
        # rule for computing dnm = (dn1 + dn2)/2 is used, then the error increases
        # by a factor of 2.)  Setting this equal to epsilon gives sig12 = etol2.
        # Here 0.1 is a safety factor (error decreased by 100) and max(0.001,
        # abs(f)) stops etol2 getting too large in the nearly spherical case.
        self._etol2 = 0.1 * Geodesic.tol2_ / math.sqrt( max(0.001, abs(self.f)) *
                                                    min(1.0, 1-self.f/2) / 2 )
        if not(math.isfinite(self.a) and self.a > 0):
            raise ValueError("Equatorial radius is not positive")
        if not(math.isfinite(self._b) and self._b > 0):
            raise ValueError("Polar semi-axis is not positive")
        self._A3x = list(range(Geodesic.nA3x_))
        self._C3x = list(range(Geodesic.nC3x_))
        self._C4x = list(range(Geodesic.nC4x_))
        self._A3coeff()
        self._C3coeff()
        self._C4coeff()

    def _A3coeff(self):
        """Private: return coefficients for A3"""
        coeff = [
        -3, 128,
      -2, -3, 64,
      -1, -3, -1, 16,
      3, -1, -2, 8,
      1, -1, 2,
      1, 1,
    ]
        o = 0; k = 0
        for j in range(Geodesic.nA3_ - 1, -1, -1): # coeff of eps^j
            m = min(Geodesic.nA3_ - j - 1, j) # order of polynomial in n
            self._A3x[k] = Math.polyval(m, coeff, o, self._n) / coeff[o + m + 1]
            k += 1
            o += m + 2

    def _C3coeff(self):
        """Private: return coefficients for C3"""
        coeff = [
        3, 128,
      2, 5, 128,
      -1, 3, 3, 64,
      -1, 0, 1, 8,
      -1, 1, 4,
      5, 256,
      1, 3, 128,
      -3, -2, 3, 64,
      1, -3, 2, 32,
      7, 512,
      -10, 9, 384,
      5, -9, 5, 192,
      7, 512,
      -14, 7, 512,
      21, 2560,
    ]
        o = 0; k = 0
        for l in range(1, Geodesic.nC3_): # l is index of C3[l]
            for j in range(Geodesic.nC3_ - 1, l - 1, -1): # coeff of eps^j
                m = min(Geodesic.nC3_ - j - 1, j) # order of polynomial in n
                self._C3x[k] = Math.polyval(m, coeff, o, self._n) / coeff[o + m + 1]
                k += 1
                o += m + 2

    def _C4coeff(self):
        """Private: return coefficients for C4"""
        coeff = [
        97, 15015,
      1088, 156, 45045,
      -224, -4784, 1573, 45045,
      -10656, 14144, -4576, -858, 45045,
      64, 624, -4576, 6864, -3003, 15015,
      100, 208, 572, 3432, -12012, 30030, 45045,
      1, 9009,
      -2944, 468, 135135,
      5792, 1040, -1287, 135135,
      5952, -11648, 9152, -2574, 135135,
      -64, -624, 4576, -6864, 3003, 135135,
      8, 10725,
      1856, -936, 225225,
      -8448, 4992, -1144, 225225,
      -1440, 4160, -4576, 1716, 225225,
      -136, 63063,
      1024, -208, 105105,
      3584, -3328, 1144, 315315,
      -128, 135135,
      -2560, 832, 405405,
      128, 99099,
    ]
        o = 0; k = 0
        for l in range(Geodesic.nC4_): # l is index of C4[l]
            for j in range(Geodesic.nC4_ - 1, l - 1, -1): # coeff of eps^j
                m = Geodesic.nC4_ - j - 1 # order of polynomial in n
                self._C4x[k] = Math.polyval(m, coeff, o, self._n) / coeff[o + m + 1]
                k += 1
                o += m + 2

    def _A3f(self, eps):
        """Private: return A3"""
        # Evaluate A3
        return Math.polyval(Geodesic.nA3_ - 1, self._A3x, 0, eps)

    def _C3f(self, eps, c):
        """Private: return C3"""
        # Evaluate C3
        # Elements c[1] thru c[nC3_ - 1] are set
        mult = 1
        o = 0
        for l in range(1, Geodesic.nC3_): # l is index of C3[l]
            m = Geodesic.nC3_ - l - 1       # order of polynomial in eps
            mult *= eps
            c[l] = mult * Math.polyval(m, self._C3x, o, eps)
            o += m + 1 

    # return s12b, m12b, m0, M12, M21
    def _Lengths(self, eps, sig12,
               ssig1, csig1, dn1, ssig2, csig2, dn2, cbet1, cbet2, outmask,
               # Scratch areas of the right size
               C1a, C2a):
        """Private: return a bunch of lengths"""
        # Return s12b, m12b, m0, M12, M21, where
        # m12b = (reduced length)/_b; s12b = distance/_b,
        # m0 = coefficient of secular term in expression for reduced length.
        outmask &= Geodesic.OUT_MASK
        # outmask & DISTANCE: set s12b
        # outmask & REDUCEDLENGTH: set m12b & m0
        # outmask & GEODESICSCALE: set M12 & M21

        s12b = m12b = m0 = M12 = M21 = math.nan
        if outmask & (Geodesic.DISTANCE | Geodesic.REDUCEDLENGTH |
                  Geodesic.GEODESICSCALE):
            A1 = Geodesic._A1m1f(eps)
            Geodesic._C1f(eps, C1a)
            if outmask & (Geodesic.REDUCEDLENGTH | Geodesic.GEODESICSCALE):
                A2 = Geodesic._A2m1f(eps)
                Geodesic._C2f(eps, C2a)
                m0x = A1 - A2
                A2 = 1 + A2
            A1 = 1 + A1
        if outmask & Geodesic.DISTANCE:
            B1 = (Geodesic._SinCosSeries(True, ssig2, csig2, C1a) -
            Geodesic._SinCosSeries(True, ssig1, csig1, C1a))
            # Missing a factor of _b
            s12b = A1 * (sig12 + B1)
            if outmask & (Geodesic.REDUCEDLENGTH | Geodesic.GEODESICSCALE):
                B2 = (Geodesic._SinCosSeries(True, ssig2, csig2, C2a) -
              Geodesic._SinCosSeries(True, ssig1, csig1, C2a))
                J12 = m0x * sig12 + (A1 * B1 - A2 * B2)
        elif outmask & (Geodesic.REDUCEDLENGTH | Geodesic.GEODESICSCALE):
            # Assume here that nC1_ >= nC2_
            for l in range(1, Geodesic.nC2_):
                C2a[l] = A1 * C1a[l] - A2 * C2a[l]
            J12 = m0x * sig12 + (Geodesic._SinCosSeries(True, ssig2, csig2, C2a) -
                           Geodesic._SinCosSeries(True, ssig1, csig1, C2a))
        if outmask & Geodesic.REDUCEDLENGTH:
            m0 = m0x
            # Missing a factor of _b.
            # Add parens around (csig1 * ssig2) and (ssig1 * csig2) to ensure
            # accurate cancellation in the case of coincident points.
            m12b = (dn2 * (csig1 * ssig2) - dn1 * (ssig1 * csig2) -
              csig1 * csig2 * J12)
        if outmask & Geodesic.GEODESICSCALE:
            csig12 = csig1 * csig2 + ssig1 * ssig2
            t = self._ep2 * (cbet1 - cbet2) * (cbet1 + cbet2) / (dn1 + dn2)
            M12 = csig12 + (t * ssig2 - csig2 * J12) * ssig1 / dn1
            M21 = csig12 - (t * ssig1 - csig1 * J12) * ssig2 / dn2
        return s12b, m12b, m0, M12, M21

    # return sig12, salp1, calp1, salp2, calp2, dnm
    def _InverseStart(self, sbet1, cbet1, dn1, sbet2, cbet2, dn2,
                    lam12, slam12, clam12,
                    # Scratch areas of the right size
                    C1a, C2a):
        """Private: Find a starting value for Newton's method."""
        # Return a starting point for Newton's method in salp1 and calp1 (function
        # value is -1).  If Newton's method doesn't need to be used, return also
        # salp2 and calp2 and function value is sig12.
        sig12 = -1; salp2 = calp2 = dnm = math.nan # Return values
        # bet12 = bet2 - bet1 in [0, pi); bet12a = bet2 + bet1 in (-pi, 0]
        sbet12 = sbet2 * cbet1 - cbet2 * sbet1
        cbet12 = cbet2 * cbet1 + sbet2 * sbet1
        # Volatile declaration needed to fix inverse cases
        # 88.202499451857 0 -88.202499451857 179.981022032992859592
        # 89.262080389218 0 -89.262080389218 179.992207982775375662
        # 89.333123580033 0 -89.333123580032997687 179.99295812360148422
        # which otherwise fail with g++ 4.4.4 x86 -O3
        sbet12a = sbet2 * cbet1
        sbet12a += cbet2 * sbet1

        shortline = cbet12 >= 0 and sbet12 < 0.5 and cbet2 * lam12 < 0.5
        if shortline:
            sbetm2 = Math.sq(sbet1 + sbet2)
            # sin((bet1+bet2)/2)^2
            # =  (sbet1 + sbet2)^2 / ((sbet1 + sbet2)^2 + (cbet1 + cbet2)^2)
            sbetm2 /= sbetm2 + Math.sq(cbet1 + cbet2)
            dnm = math.sqrt(1 + self._ep2 * sbetm2)
            omg12 = lam12 / (self._f1 * dnm)
            somg12 = math.sin(omg12); comg12 = math.cos(omg12)
        else:
            somg12 = slam12; comg12 = clam12

        salp1 = cbet2 * somg12
        calp1 = (
        sbet12 + cbet2 * sbet1 * Math.sq(somg12) / (1 + comg12) if comg12 >= 0
      else sbet12a - cbet2 * sbet1 * Math.sq(somg12) / (1 - comg12))

        ssig12 = math.hypot(salp1, calp1)
        csig12 = sbet1 * sbet2 + cbet1 * cbet2 * comg12

        if shortline and ssig12 < self._etol2:
            # really short lines
            salp2 = cbet1 * somg12
            calp2 = sbet12 - cbet1 * sbet2 * (Math.sq(somg12) / (1 + comg12)
                                        if comg12 >= 0 else 1 - comg12)
            salp2, calp2 = Math.norm(salp2, calp2)
            # Set return value
            sig12 = math.atan2(ssig12, csig12)
        elif (abs(self._n) >= 0.1 or # Skip astroid calc if too eccentric
          csig12 >= 0 or
          ssig12 >= 6 * abs(self._n) * math.pi * Math.sq(cbet1)):
            # Nothing to do, zeroth order spherical approximation is OK
            pass
        else:
            # Scale lam12 and bet2 to x, y coordinate system where antipodal point
            # is at origin and singular point is at y = 0, x = -1.
            # real y, lamscale, betscale
            lam12x = math.atan2(-slam12, -clam12)
            if self.f >= 0:            # In fact f == 0 does not get here
                # x = dlong, y = dlat
                k2 = Math.sq(sbet1) * self._ep2
                eps = k2 / (2 * (1 + math.sqrt(1 + k2)) + k2)
                lamscale = self.f * cbet1 * self._A3f(eps) * math.pi
                betscale = lamscale * cbet1
                x = lam12x / lamscale
                y = sbet12a / betscale
            else:                     # _f < 0
                # x = dlat, y = dlong
                cbet12a = cbet2 * cbet1 - sbet2 * sbet1
                bet12a = math.atan2(sbet12a, cbet12a)
                # real m12b, m0, dummy
                # In the case of lon12 = 180, this repeats a calculation made in
                # Inverse.
                dummy, m12b, m0, dummy, dummy = self._Lengths(
            self._n, math.pi + bet12a, sbet1, -cbet1, dn1, sbet2, cbet2, dn2,
          cbet1, cbet2, Geodesic.REDUCEDLENGTH, C1a, C2a)
                x = -1 + m12b / (cbet1 * cbet2 * m0 * math.pi)
                betscale = (sbet12a / x if x < -0.01
                    else -self.f * Math.sq(cbet1) * math.pi)
                lamscale = betscale / cbet1
                y = lam12x / lamscale

            if y > -Geodesic.tol1_ and x > -1 - Geodesic.xthresh_:
                # strip near cut
                if self.f >= 0:
                    salp1 = min(1.0, -x); calp1 = - math.sqrt(1 - Math.sq(salp1))
                else:
                    calp1 = max((0.0 if x > -Geodesic.tol1_ else -1.0), x)
                    salp1 = math.sqrt(1 - Math.sq(calp1))
            else:
                # Estimate alp1, by solving the astroid problem.
                #
                # Could estimate alpha1 = theta + pi/2, directly, i.e.,
                #   calp1 = y/k; salp1 = -x/(1+k);  for _f >= 0
                #   calp1 = x/(1+k); salp1 = -y/k;  for _f < 0 (need to check)
                #
                # However, it's better to estimate omg12 from astroid and use
                # spherical formula to compute alp1.  This reduces the mean number of
                # Newton iterations for astroid cases from 2.24 (min 0, max 6) to 2.12
                # (min 0 max 5).  The changes in the number of iterations are as
                # follows:
                #
                # change percent
                #    1       5
                #    0      78
                #   -1      16
                #   -2       0.6
                #   -3       0.04
                #   -4       0.002
                #
                # The histogram of iterations is (m = number of iterations estimating
                # alp1 directly, n = number of iterations estimating via omg12, total
                # number of trials = 148605):
                #
                #  iter    m      n
                #    0   148    186
                #    1 13046  13845
                #    2 93315 102225
                #    3 36189  32341
                #    4  5396      7
                #    5   455      1
                #    6    56      0
                #
                # Because omg12 is near pi, estimate work with omg12a = pi - omg12
                k = Geodesic._Astroid(x, y)
                omg12a = lamscale * ( -x * k/(1 + k) if self.f >= 0
                              else -y * (1 + k)/k )
                somg12 = math.sin(omg12a); comg12 = -math.cos(omg12a)
                # Update spherical estimate of alp1 using omg12 instead of lam12
                salp1 = cbet2 * somg12
                calp1 = sbet12a - cbet2 * sbet1 * Math.sq(somg12) / (1 - comg12)
        # Sanity check on starting guess.  Backwards check allows NaN through.
        if not (salp1 <= 0):
            salp1, calp1 = Math.norm(salp1, calp1)
        else:
            salp1 = 1; calp1 = 0
        return sig12, salp1, calp1, salp2, calp2, dnm

    # return lam12, salp2, calp2, sig12, ssig1, csig1, ssig2, csig2, eps,
    # domg12, dlam12
    
    def _Lambda12(self, sbet1, cbet1, dn1, sbet2, cbet2, dn2, salp1, calp1,
                slam120, clam120, diffp,
                # Scratch areas of the right size
                C1a, C2a, C3a):
        """Private: Solve hybrid problem"""
        if sbet1 == 0 and calp1 == 0:
            # Break degeneracy of equatorial line.  This case has already been
            # handled.
            calp1 = -Geodesic.tiny_

        # sin(alp1) * cos(bet1) = sin(alp0)
        salp0 = salp1 * cbet1
        calp0 = math.hypot(calp1, salp1 * sbet1) # calp0 > 0

        # real somg1, comg1, somg2, comg2, lam12
        # tan(bet1) = tan(sig1) * cos(alp1)
        # tan(omg1) = sin(alp0) * tan(sig1) = tan(omg1)=tan(alp1)*sin(bet1)
        ssig1 = sbet1; somg1 = salp0 * sbet1
        csig1 = comg1 = calp1 * cbet1
        ssig1, csig1 = Math.norm(ssig1, csig1)
        # Math.norm(somg1, comg1); -- don't need to normalize!

        # Enforce symmetries in the case abs(bet2) = -bet1.  Need to be careful
        # about this case, since this can yield singularities in the Newton
        # iteration.
        # sin(alp2) * cos(bet2) = sin(alp0)
        salp2 = salp0 / cbet2 if cbet2 != cbet1 else salp1
        # calp2 = sqrt(1 - sq(salp2))
        #       = sqrt(sq(calp0) - sq(sbet2)) / cbet2
        # and subst for calp0 and rearrange to give (choose positive sqrt
        # to give alp2 in [0, pi/2]).
        calp2 = (math.sqrt(Math.sq(calp1 * cbet1) +
                       ((cbet2 - cbet1) * (cbet1 + cbet2) if cbet1 < -sbet1
                        else (sbet1 - sbet2) * (sbet1 + sbet2))) / cbet2
             if cbet2 != cbet1 or abs(sbet2) != -sbet1 else abs(calp1))
        # tan(bet2) = tan(sig2) * cos(alp2)
        # tan(omg2) = sin(alp0) * tan(sig2).
        ssig2 = sbet2; somg2 = salp0 * sbet2
        csig2 = comg2 = calp2 * cbet2
        ssig2, csig2 = Math.norm(ssig2, csig2)
        # Math.norm(somg2, comg2); -- don't need to normalize!

        # sig12 = sig2 - sig1, limit to [0, pi]
        sig12 = math.atan2(max(0.0, csig1 * ssig2 - ssig1 * csig2) + 0.0,
                       csig1 * csig2 + ssig1 * ssig2)

        # omg12 = omg2 - omg1, limit to [0, pi]
        somg12 = max(0.0, comg1 * somg2 - somg1 * comg2) + 0.0
        comg12 =          comg1 * comg2 + somg1 * somg2
        # eta = omg12 - lam120
        eta = math.atan2(somg12 * clam120 - comg12 * slam120,
                     comg12 * clam120 + somg12 * slam120)

        # real B312
        k2 = Math.sq(calp0) * self._ep2
        eps = k2 / (2 * (1 + math.sqrt(1 + k2)) + k2)
        self._C3f(eps, C3a)
        B312 = (Geodesic._SinCosSeries(True, ssig2, csig2, C3a) -
            Geodesic._SinCosSeries(True, ssig1, csig1, C3a))
        domg12 =  -self.f * self._A3f(eps) * salp0 * (sig12 + B312)
        lam12 = eta + domg12

        if diffp:
            if calp2 == 0:
                dlam12 = - 2 * self._f1 * dn1 / sbet1
            else:
                dummy, dlam12, dummy, dummy, dummy = self._Lengths(
            eps, sig12, ssig1, csig1, dn1, ssig2, csig2, dn2, cbet1, cbet2,
          Geodesic.REDUCEDLENGTH, C1a, C2a)
                dlam12 *= self._f1 / (calp2 * cbet2)
        else:
            dlam12 = math.nan

        return (lam12, salp2, calp2, sig12, ssig1, csig1, ssig2, csig2, eps,
            domg12, dlam12)

    # return a12, s12, salp1, calp1, salp2, calp2, m12, M12, M21, S12
    def _GenInverse(self, lat1, lon1, lat2, lon2, outmask):
        """Private: General version of the inverse problem"""
        a12 = s12 = m12 = M12 = M21 = S12 = math.nan # return vals

        outmask &= Geodesic.OUT_MASK
        # Compute longitude difference (AngDiff does this carefully).  Result is
        # in [-180, 180] but -180 is only for west-going geodesics.  180 is for
        # east-going and meridional geodesics.
        lon12, lon12s = Math.AngDiff(lon1, lon2)
        # Make longitude difference positive.
        lonsign = math.copysign(1, lon12)
        lon12 = lonsign * lon12; lon12s = lonsign * lon12s
        lam12 = math.radians(lon12)
        # Calculate sincos of lon12 + error (this applies AngRound internally).
        slam12, clam12 = Math.sincosde(lon12, lon12s)
        lon12s = (180 - lon12) - lon12s # the supplementary longitude difference

        # If really close to the equator, treat as on equator.
        lat1 = Math.AngRound(Math.LatFix(lat1))
        lat2 = Math.AngRound(Math.LatFix(lat2))
        # Swap points so that point with higher (abs) latitude is point 1
        # If one latitude is a nan, then it becomes lat1.
        swapp = -1 if abs(lat1) < abs(lat2) or math.isnan(lat2) else 1
        if swapp < 0:
            lonsign *= -1
            lat2, lat1 = lat1, lat2
        # Make lat1 <= 0
        latsign = math.copysign(1, -lat1)
        lat1 *= latsign
        lat2 *= latsign
        # Now we have
        #
        #     0 <= lon12 <= 180
        #     -90 <= lat1 <= 0
        #     lat1 <= lat2 <= -lat1
        #
        # longsign, swapp, latsign register the transformation to bring the
        # coordinates to this canonical form.  In all cases, 1 means no change was
        # made.  We make these transformations so that there are few cases to
        # check, e.g., on verifying quadrants in atan2.  In addition, this
        # enforces some symmetries in the results returned.

        # real phi, sbet1, cbet1, sbet2, cbet2, s12x, m12x

        sbet1, cbet1 = Math.sincosd(lat1); sbet1 *= self._f1
        # Ensure cbet1 = +epsilon at poles
        sbet1, cbet1 = Math.norm(sbet1, cbet1); cbet1 = max(Geodesic.tiny_, cbet1)

        sbet2, cbet2 = Math.sincosd(lat2); sbet2 *= self._f1
        # Ensure cbet2 = +epsilon at poles
        sbet2, cbet2 = Math.norm(sbet2, cbet2); cbet2 = max(Geodesic.tiny_, cbet2)

        # If cbet1 < -sbet1, then cbet2 - cbet1 is a sensitive measure of the
        # |bet1| - |bet2|.  Alternatively (cbet1 >= -sbet1), abs(sbet2) + sbet1 is
        # a better measure.  This logic is used in assigning calp2 in Lambda12.
        # Sometimes these quantities vanish and in that case we force bet2 = +/-
        # bet1 exactly.  An example where is is necessary is the inverse problem
        # 48.522876735459 0 -48.52287673545898293 179.599720456223079643
        # which failed with Visual Studio 10 (Release and Debug)

        if cbet1 < -sbet1:
            if cbet2 == cbet1:
                sbet2 = math.copysign(sbet1, sbet2)
        else:
            if abs(sbet2) == -sbet1:
                cbet2 = cbet1

        dn1 = math.sqrt(1 + self._ep2 * Math.sq(sbet1))
        dn2 = math.sqrt(1 + self._ep2 * Math.sq(sbet2))

        # real a12, sig12, calp1, salp1, calp2, salp2
        # index zero elements of these arrays are unused
        C1a = list(range(Geodesic.nC1_ + 1))
        C2a = list(range(Geodesic.nC2_ + 1))
        C3a = list(range(Geodesic.nC3_))

        meridian = lat1 == -90 or slam12 == 0

        if meridian:

            # Endpoints are on a single full meridian, so the geodesic might lie on
            # a meridian.

            calp1 = clam12; salp1 = slam12 # Head to the target longitude
            calp2 = 1.0; salp2 = 0.0       # At the target we're heading north

            # tan(bet) = tan(sig) * cos(alp)
            ssig1 = sbet1; csig1 = calp1 * cbet1
            ssig2 = sbet2; csig2 = calp2 * cbet2

            # sig12 = sig2 - sig1
            sig12 = math.atan2(max(0.0, csig1 * ssig2 - ssig1 * csig2) + 0.0,
                         csig1 * csig2 + ssig1 * ssig2)

            s12x, m12x, dummy, M12, M21 = self._Lengths(
          self._n, sig12, ssig1, csig1, dn1, ssig2, csig2, dn2, cbet1, cbet2,
        outmask | Geodesic.DISTANCE | Geodesic.REDUCEDLENGTH, C1a, C2a)

            # Add the check for sig12 since zero length geodesics might yield m12 <
            # 0.  Test case was
            #
            #    echo 20.001 0 20.001 0 | GeodSolve -i
            #
            # In fact, we will have sig12 > pi/2 for meridional geodesic which is
            # not a shortest path.
            if sig12 < 1 or m12x >= 0:
                if (sig12 < 3 * Geodesic.tiny_ or
            # Prevent negative s12 or m12 for short lines
            (sig12 < Geodesic.tol0_ and (s12x < 0 or m12x < 0))):
                    sig12 = m12x = s12x = 0.0
                m12x *= self._b
                s12x *= self._b
                a12 = math.degrees(sig12)
            else:
                # m12 < 0, i.e., prolate and too close to anti-podal
                meridian = False
        # end if meridian:

        # somg12 == 2 marks that it needs to be calculated
        somg12 = 2.0; comg12 = 0.0; omg12 = 0.0
        if (not meridian and
        sbet1 == 0 and   # and sbet2 == 0
        # Mimic the way Lambda12 works with calp1 = 0
        (self.f <= 0 or lon12s >= self.f * 180)):

            # Geodesic runs along equator
            calp1 = calp2 = 0.0; salp1 = salp2 = 1.0
            s12x = self.a * lam12
            sig12 = omg12 = lam12 / self._f1
            m12x = self._b * math.sin(sig12)
            if outmask & Geodesic.GEODESICSCALE:
                M12 = M21 = math.cos(sig12)
            a12 = lon12 / self._f1

        elif not meridian:

            # Now point1 and point2 belong within a hemisphere bounded by a
            # meridian and geodesic is neither meridional or equatorial.

            # Figure a starting point for Newton's method
            sig12, salp1, calp1, salp2, calp2, dnm = self._InverseStart(
          sbet1, cbet1, dn1, sbet2, cbet2, dn2, lam12, slam12, clam12, C1a, C2a)

            if sig12 >= 0:
                # Short lines (InverseStart sets salp2, calp2, dnm)
                s12x = sig12 * self._b * dnm
                m12x = (Math.sq(dnm) * self._b * math.sin(sig12 / dnm))
                if outmask & Geodesic.GEODESICSCALE:
                    M12 = M21 = math.cos(sig12 / dnm)
                a12 = math.degrees(sig12)
                omg12 = lam12 / (self._f1 * dnm)
            else:

                # Newton's method.  This is a straightforward solution of f(alp1) =
                # lambda12(alp1) - lam12 = 0 with one wrinkle.  f(alp) has exactly one
                # root in the interval (0, pi) and its derivative is positive at the
                # root.  Thus f(alp) is positive for alp > alp1 and negative for alp <
                # alp1.  During the course of the iteration, a range (alp1a, alp1b) is
                # maintained which brackets the root and with each evaluation of f(alp)
                # the range is shrunk if possible.  Newton's method is restarted
                # whenever the derivative of f is negative (because the new value of
                # alp1 is then further from the solution) or if the new estimate of
                # alp1 lies outside (0,pi); in this case, the new starting guess is
                # taken to be (alp1a + alp1b) / 2.
                # real ssig1, csig1, ssig2, csig2, eps
                numit = 0
                tripn = tripb = False
                # Bracketing range
                salp1a = Geodesic.tiny_; calp1a = 1.0
                salp1b = Geodesic.tiny_; calp1b = -1.0

                while numit < Geodesic.maxit2_:
                    # the WGS84 test set: mean = 1.47, sd = 1.25, max = 16
                    # WGS84 and random input: mean = 2.85, sd = 0.60
                    (v, salp2, calp2, sig12, ssig1, csig1, ssig2, csig2,
           eps, domg12, dv) = self._Lambda12(
               sbet1, cbet1, dn1, sbet2, cbet2, dn2,
             salp1, calp1, slam12, clam12, numit < Geodesic.maxit1_,
             C1a, C2a, C3a)
                    # Reversed test to allow escape with NaNs
                    if tripb or not (abs(v) >= (8 if tripn else 1) * Geodesic.tol0_):
                        break
                    # Update bracketing values
                    if v > 0 and (numit > Geodesic.maxit1_ or
                        calp1/salp1 > calp1b/salp1b):
                        salp1b = salp1; calp1b = calp1
                    elif v < 0 and (numit > Geodesic.maxit1_ or
                          calp1/salp1 < calp1a/salp1a):
                        salp1a = salp1; calp1a = calp1

                    numit += 1
                    if numit < Geodesic.maxit1_ and dv > 0:
                        dalp1 = -v/dv
                        sdalp1 = math.sin(dalp1); cdalp1 = math.cos(dalp1)
                        nsalp1 = salp1 * cdalp1 + calp1 * sdalp1
                        if nsalp1 > 0 and abs(dalp1) < math.pi:
                            calp1 = calp1 * cdalp1 - salp1 * sdalp1
                            salp1 = nsalp1
                            salp1, calp1 = Math.norm(salp1, calp1)
                            # In some regimes we don't get quadratic convergence because
                            # slope -> 0.  So use convergence conditions based on epsilon
                            # instead of sqrt(epsilon).
                            tripn = abs(v) <= 16 * Geodesic.tol0_
                            continue
                    # Either dv was not positive or updated value was outside
                    # legal range.  Use the midpoint of the bracket as the next
                    # estimate.  This mechanism is not needed for the WGS84
                    # ellipsoid, but it does catch problems with more eccentric
                    # ellipsoids.  Its efficacy is such for
                    # the WGS84 test set with the starting guess set to alp1 = 90deg:
                    # the WGS84 test set: mean = 5.21, sd = 3.93, max = 24
                    # WGS84 and random input: mean = 4.74, sd = 0.99
                    salp1 = (salp1a + salp1b)/2
                    calp1 = (calp1a + calp1b)/2
                    salp1, calp1 = Math.norm(salp1, calp1)
                    tripn = False
                    tripb = (abs(salp1a - salp1) + (calp1a - calp1) < Geodesic.tolb_ or
                   abs(salp1 - salp1b) + (calp1 - calp1b) < Geodesic.tolb_)

                lengthmask = (outmask |
                      (Geodesic.DISTANCE
                       if (outmask & (Geodesic.REDUCEDLENGTH |
                                      Geodesic.GEODESICSCALE))
                       else Geodesic.EMPTY))
                s12x, m12x, dummy, M12, M21 = self._Lengths(
            eps, sig12, ssig1, csig1, dn1, ssig2, csig2, dn2, cbet1, cbet2,
          lengthmask, C1a, C2a)

                m12x *= self._b
                s12x *= self._b
                a12 = math.degrees(sig12)
                if outmask & Geodesic.AREA:
                    # omg12 = lam12 - domg12
                    sdomg12 = math.sin(domg12); cdomg12 = math.cos(domg12)
                    somg12 = slam12 * cdomg12 - clam12 * sdomg12
                    comg12 = clam12 * cdomg12 + slam12 * sdomg12

        # end elif not meridian

        if outmask & Geodesic.DISTANCE:
            s12 = 0.0 + s12x          # Convert -0 to 0

        if outmask & Geodesic.REDUCEDLENGTH:
            m12 = 0.0 + m12x          # Convert -0 to 0

        if outmask & Geodesic.AREA:
            # From Lambda12: sin(alp1) * cos(bet1) = sin(alp0)
            salp0 = salp1 * cbet1
            calp0 = math.hypot(calp1, salp1 * sbet1) # calp0 > 0
            # real alp12
            if calp0 != 0 and salp0 != 0:
                # From Lambda12: tan(bet) = tan(sig) * cos(alp)
                ssig1 = sbet1; csig1 = calp1 * cbet1
                ssig2 = sbet2; csig2 = calp2 * cbet2
                k2 = Math.sq(calp0) * self._ep2
                eps = k2 / (2 * (1 + math.sqrt(1 + k2)) + k2)
                # Multiplier = a^2 * e^2 * cos(alpha0) * sin(alpha0).
                A4 = Math.sq(self.a) * calp0 * salp0 * self._e2
                ssig1, csig1 = Math.norm(ssig1, csig1)
                ssig2, csig2 = Math.norm(ssig2, csig2)
                C4a = list(range(Geodesic.nC4_))
                self._C4f(eps, C4a)
                B41 = Geodesic._SinCosSeries(False, ssig1, csig1, C4a)
                B42 = Geodesic._SinCosSeries(False, ssig2, csig2, C4a)
                S12 = A4 * (B42 - B41)
            else:
                # Avoid problems with indeterminate sig1, sig2 on equator
                S12 = 0.0

            if not meridian and somg12 == 2.0:
                somg12 = math.sin(omg12); comg12 = math.cos(omg12)

            if (not meridian and
          # omg12 < 3/4 * pi
          comg12 > -0.7071 and   # Long difference not too big
          sbet2 - sbet1 < 1.75): # Lat difference not too big
                # Use tan(Gamma/2) = tan(omg12/2)
                # * (tan(bet1/2)+tan(bet2/2))/(1+tan(bet1/2)*tan(bet2/2))
                # with tan(x/2) = sin(x)/(1+cos(x))
                domg12 = 1 + comg12; dbet1 = 1 + cbet1; dbet2 = 1 + cbet2
                alp12 = 2 * math.atan2( somg12 * ( sbet1 * dbet2 + sbet2 * dbet1 ),
                                domg12 * ( sbet1 * sbet2 + dbet1 * dbet2 ) )
            else:
                # alp12 = alp2 - alp1, used in atan2 so no need to normalize
                salp12 = salp2 * calp1 - calp2 * salp1
                calp12 = calp2 * calp1 + salp2 * salp1
                # The right thing appears to happen if alp1 = +/-180 and alp2 = 0, viz
                # salp12 = -0 and alp12 = -180.  However this depends on the sign
                # being attached to 0 correctly.  The following ensures the correct
                # behavior.
                if salp12 == 0 and calp12 < 0:
                    salp12 = Geodesic.tiny_ * calp1
                    calp12 = -1.0
                alp12 = math.atan2(salp12, calp12)
            S12 += self._c2 * alp12
            S12 *= swapp * lonsign * latsign
            # Convert -0 to 0
            S12 += 0.0

        # Convert calp, salp to azimuth accounting for lonsign, swapp, latsign.
        if swapp < 0:
            salp2, salp1 = salp1, salp2
            calp2, calp1 = calp1, calp2
            if outmask & Geodesic.GEODESICSCALE:
                M21, M12 = M12, M21

        salp1 *= swapp * lonsign; calp1 *= swapp * latsign
        salp2 *= swapp * lonsign; calp2 *= swapp * latsign

        return a12, s12, salp1, calp1, salp2, calp2, m12, M12, M21, S12

    def Inverse(self, lat1, lon1, lat2, lon2,
              outmask = GeodesicCapability.STANDARD):
        """Solve the inverse geodesic problem

        :param lat1: latitude of the first point in degrees
        :param lon1: longitude of the first point in degrees
        :param lat2: latitude of the second point in degrees
        :param lon2: longitude of the second point in degrees
        :param outmask: the :ref:`output mask <outmask>`
        :return: a :ref:`dict`

        Compute geodesic between (*lat1*, *lon1*) and (*lat2*, *lon2*).
        The default value of *outmask* is STANDARD, i.e., the *lat1*,
        *lon1*, *azi1*, *lat2*, *lon2*, *azi2*, *s12*, *a12* entries are
        returned.

        """

        a12, s12, salp1,calp1, salp2,calp2, m12, M12, M21, S12 = self._GenInverse(
        lat1, lon1, lat2, lon2, outmask)
        outmask &= Geodesic.OUT_MASK
        if outmask & Geodesic.LONG_UNROLL:
            lon12, e = Math.AngDiff(lon1, lon2)
            lon2 = (lon1 + lon12) + e
        else:
            lon2 = Math.AngNormalize(lon2)
        result = {'lat1': Math.LatFix(lat1),
              'lon1': lon1 if outmask & Geodesic.LONG_UNROLL else
              Math.AngNormalize(lon1),
              'lat2': Math.LatFix(lat2),
              'lon2': lon2}
        result['a12'] = a12
        if outmask & Geodesic.DISTANCE: result['s12'] = s12
        if outmask & Geodesic.AZIMUTH:
            result['azi1'] = Math.atan2d(salp1, calp1)
            result['azi2'] = Math.atan2d(salp2, calp2)
        if outmask & Geodesic.REDUCEDLENGTH: result['m12'] = m12
        if outmask & Geodesic.GEODESICSCALE:
            result['M12'] = M12; result['M21'] = M21
        if outmask & Geodesic.AREA: result['S12'] = S12
        return result

    # return a12, lat2, lon2, azi2, s12, m12, M12, M21, S12
    def _GenDirect(self, lat1, lon1, azi1, arcmode, s12_a12, outmask):
        """Private: General version of direct problem"""
        from geographiclib.geodesicline import GeodesicLine
        # Automatically supply DISTANCE_IN if necessary
        if not arcmode: outmask |= Geodesic.DISTANCE_IN
        line = GeodesicLine(self, lat1, lon1, azi1, outmask)
        return line._GenPosition(arcmode, s12_a12, outmask)
 

    EMPTY         = GeodesicCapability.EMPTY
    """No capabilities, no output."""
    LATITUDE      = GeodesicCapability.LATITUDE
    """Calculate latitude *lat2*."""
    LONGITUDE     = GeodesicCapability.LONGITUDE
    """Calculate longitude *lon2*."""
    AZIMUTH       = GeodesicCapability.AZIMUTH
    """Calculate azimuths *azi1* and *azi2*."""
    DISTANCE      = GeodesicCapability.DISTANCE
    """Calculate distance *s12*."""
    STANDARD      = GeodesicCapability.STANDARD
    """All of the above."""
    DISTANCE_IN   = GeodesicCapability.DISTANCE_IN
    """Allow distance *s12* to be used as input in the direct geodesic
  problem."""
    REDUCEDLENGTH = GeodesicCapability.REDUCEDLENGTH
    """Calculate reduced length *m12*."""
    GEODESICSCALE = GeodesicCapability.GEODESICSCALE
    """Calculate geodesic scales *M12* and *M21*."""
    AREA          = GeodesicCapability.AREA
    """Calculate area *S12*."""
    ALL           = GeodesicCapability.ALL
    """All of the above."""
    LONG_UNROLL   = GeodesicCapability.LONG_UNROLL
    """Unroll longitudes, rather than reducing them to the range
  [-180d,180d].

  """

Geodesic.WGS84 = Geodesic(Constants.WGS84_a, Constants.WGS84_f)
"""Instantiation for the WGS84 ellipsoid"""

class Distance:
    """ Modified to remve unit conversions to stay with RCAIDE conventions - Oct. 2024
    Base class for other distance algorithms. Represents a distance.

    Can be used for units conversion::

        >>> from geopy.distance import Distance
        >>> Distance(miles=10).km
        16.09344

    Distance instances have all *distance* properties from :mod:`geopy.units`,
    e.g.: ``km``, ``m``, ``meters``, ``miles`` and so on.

    Distance instances are immutable.

    They support comparison::

        >>> from geopy.distance import Distance
        >>> Distance(kilometers=2) == Distance(meters=2000)
        True
        >>> Distance(kilometers=2) > Distance(miles=1)
        True

    String representation::

        >>> from geopy.distance import Distance
        >>> repr(Distance(kilometers=2))
        'Distance(2.0)'
        >>> str(Distance(kilometers=2))
        '2.0 km'
        >>> repr(Distance(miles=2))
        'Distance(3.218688)'
        >>> str(Distance(miles=2))
        '3.218688 km'

    Arithmetics::

        >>> from geopy.distance import Distance
        >>> -Distance(miles=2)
        Distance(-3.218688)
        >>> Distance(miles=2) + Distance(kilometers=1)
        Distance(4.218688)
        >>> Distance(miles=2) - Distance(kilometers=1)
        Distance(2.218688)
        >>> Distance(kilometers=6) * 5
        Distance(30.0)
        >>> Distance(kilometers=6) / 5
        Distance(1.2)
    """
    
    def __init__(self, *args, **kwargs):
        """
        There are 3 ways to create a distance:

        - From kilometers::

            >>> from geopy.distance import Distance
            >>> Distance(1.42)
            Distance(1.42)

        - From points (for non-abstract distances only),
          calculated as a sum of distances between all points::

            >>> from geopy.distance import geodesic
            >>> geodesic((40, 160), (40.1, 160.1))
            Distance(14.003702498106215)
            >>> geodesic((40, 160), (40.1, 160.1), (40.2, 160.2))
            Distance(27.999954644813478)
        """

        kilometers = kwargs.pop('kilometers', 0)
        if len(args) == 1:
            # if we only get one argument we assume
            # it's a known distance instead of
            # calculating it first
            kilometers += args[0]
        elif len(args) > 1:
            kilometers = self.measure(args[0], args[1])
        #elif len(args) > 1:
        #    for a, b in  #util.pairwise(args):
        #kilometers +=  self.measure(a, b)

        #kilometers += units.kilometers(**kwargs)
        self.__kilometers = kilometers

    def __add__(self, other):
        if isinstance(other, Distance):
            return self.__class__(self.kilometers + other.kilometers)
        else:
            raise TypeError("Distance instance must be added with Distance instance.")

    def __neg__(self):
        return self.__class__(-self.kilometers)

    def __sub__(self, other):
        return self + -other

    def __mul__(self, other):
        if isinstance(other, Distance):
            raise TypeError("Distance instance must be multiplicated with numbers.")
        else:
            return self.__class__(self.kilometers * other)

    def __rmul__(self, other):
        if isinstance(other, Distance):
            raise TypeError("Distance instance must be multiplicated with numbers.")
        else:
            return self.__class__(other * self.kilometers)

    def __truediv__(self, other):
        if isinstance(other, Distance):
            return self.kilometers / other.kilometers
        else:
            return self.__class__(self.kilometers / other)

    def __floordiv__(self, other):
        if isinstance(other, Distance):
            return self.kilometers // other.kilometers
        else:
            return self.__class__(self.kilometers // other)

    def __abs__(self):
        return self.__class__(abs(self.kilometers))

    def __bool__(self):
        return bool(self.kilometers)

    def measure(self, a, b):
        # Intentionally not documented, because this method is not supposed
        # to be used directly.
        raise NotImplementedError("Distance is an abstract class")

    def destination(self, point, bearing, distance=None):
        """
        Calculate destination point using a starting point, bearing
        and a distance. This method works for non-abstract distances only.

        Example: a point 10 miles east from ``(34, 148)``::

            >>> import geopy.distance
            >>> geopy.distance.distance(miles=10).destination((34, 148), bearing=90)
            Point(33.99987666492774, 148.17419994321995, 0.0)

        :param point: Starting point.
        :type point: :class:`geopy.point.Point`, list or tuple of ``(latitude,
            longitude)``, or string as ``"%(latitude)s, %(longitude)s"``.

        :param float bearing: Bearing in degrees: 0 -- North, 90 -- East,
            180 -- South, 270 or -90 -- West.

        :param distance: Distance, can be used to override
            this instance::

                >>> from geopy.distance import distance, Distance
                >>> distance(miles=10).destination((34, 148), bearing=90, \distance=Distance(100))
                Point(33.995238229104764, 149.08238904409637, 0.0)

        :type distance: :class:`.Distance`

        :rtype: :class:`geopy.point.Point`
        """
        raise NotImplementedError("Distance is an abstract class")

    def __repr__(self):  # pragma: no cover
        return 'Distance(%s)' % self.kilometers

    def __str__(self):  # pragma: no cover
        return '%s km' % self.__kilometers

    def __cmp__(self, other):  # py2 only
        if isinstance(other, Distance):
            return cmp(self.kilometers, other.kilometers)
        else:
            return cmp(self.kilometers, other)

    def __hash__(self):
        return hash(self.kilometers)

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __ge__(self, other):
        return self.__cmp__(other) >= 0

    def __le__(self, other):
        return self.__cmp__(other) <= 0
    
    @property
    def kilometers(self):
        return self.__kilometers

    @property
    def km(self):
        return self.kilometers

class Geodesic_Calculate(Distance):
    """
    Calculate the geodesic distance between points.

    Set which ellipsoidal model of the earth to use by specifying an
    ``ellipsoid`` keyword argument. The default is 'WGS-84', which is the
    most globally accurate model.  If ``ellipsoid`` is a string, it is
    looked up in the `ELLIPSOIDS` dictionary to obtain the major and minor
    semiaxes and the flattening. Otherwise, it should be a tuple with those
    values.  See the comments above the `ELLIPSOIDS` dictionary for
    more information.

    Example::

        >>> from geopy.distance import geodesic
        >>> newport_ri = (41.49008, -71.312796)
        >>> cleveland_oh = (41.499498, -81.695391)
        >>> print(geodesic(newport_ri, cleveland_oh).miles)
        538.390445368

    """

    def __init__(self, *args, **kwargs):
        
        self.ellipsoid_key = None
        self.ELLIPSOID = None
        self.geod = None
        self.set_ellipsoid(kwargs.pop('ellipsoid', 'WGS-84'))
        major, minor, f = self.ELLIPSOID
        super().__init__(*args, **kwargs)

    def set_ellipsoid(self, ellipsoid):
        if isinstance(ellipsoid, str):
            try:
                self.ELLIPSOID = (6378.137, 6356.7523142, 1 / 298.257223563) # Assigns WGS-84 Ellipsoid parameters
                self.ellipsoid_key = ellipsoid
            except KeyError:
                raise Exception(
                    "Invalid ellipsoid. See geopy.distance.ELLIPSOIDS"
                )
        else:
            self.ELLIPSOID = ellipsoid
            self.ellipsoid_key = None

    def measure(self, a, b):
        #a, b = Point(a), Point(b)
        #_ensure_same_altitude(a, b)
        lat1, lon1 = a[0], a[1]
        lat2, lon2 = b[0], b[1]

        if not (isinstance(self.geod, Geodesic) and
                self.geod.a == self.ELLIPSOID[0] and
                self.geod.f == self.ELLIPSOID[2]):
            self.geod = Geodesic(self.ELLIPSOID[0], self.ELLIPSOID[2])

        s12 = self.geod.Inverse(lat1, lon1, lat2, lon2,
                                Geodesic.DISTANCE)['s12']

        return s12
    
GeodesicDistance = Geodesic_Calculate
    
# Set the default distance formula
distance = GeodesicDistance


# RCAIDE/Library/Plots/Performance/Stability/plot_lateral_stability.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_lateral_stability(results,
                             save_figure = False,
                             show_legend=True,
                             save_filename = "Lateral_Stability",
                             file_type = ".png",
                             width = 11, height = 7):
    """This plots the static stability characteristics of an aircraft 
    """  

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     
     
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height) 
    axis_1 = plt.subplot(2,2,1) 
    axis_2 = plt.subplot(2,2,2)  
    axis_3 = plt.subplot(2,2,3)    
    
    for i in range(len(results.segments)): 
        time     = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min  
        phi      = results.segments[i].conditions.aerodynamics.angles.phi[:,0] / Units.deg          
        delta_a  = results.segments[i].conditions.control_surfaces.aileron.deflection[:,0] / Units.deg  
        delta_r  = results.segments[i].conditions.control_surfaces.rudder.deflection[:,0] / Units.deg   
          
        segment_tag  =  results.segments[i].tag
        segment_name = segment_tag.replace('_', ' ')
        
        axis_1.plot(time, phi, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width, label = segment_name)
        axis_1.set_ylabel(r'$Bank Angle \phi$') 
        set_axes(axis_1)     

        axis_2.plot(time,delta_a , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_2.set_xlabel('Time (mins)')
        axis_2.set_ylabel(r'Aileron Defl. (deg)')
        set_axes(axis_2)  

        axis_3.plot(time,delta_r , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_3.set_xlabel('Time (mins)')
        axis_3.set_ylabel(r'Rudder Defl. (deg)')
        set_axes(axis_3)         
         
    if show_legend:
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})    
    
    # Adjusting the sub-plots for legend
    fig.tight_layout()
    fig.subplots_adjust(top=0.8) 
    
    # set title of plot 
    title_text    = 'Stability Coefficents'      
    fig.suptitle(title_text)
 
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return fig 


# RCAIDE/Methods/Noise/Frequency_Domain_Buildup/Rotor/LBL_VS_broadband_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# Python Package imports  
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
# Compute LBL-VS Broadband Noise 
# ---------------------------------------------------------------------------------------------------------------------- 
def LBL_VS_broadband_noise(R_c,alpha_star,delta_p,r_e,L,M,Dbar_h,f,U):
    '''This computes the laminar boundary layer compoment of broadband noise using the method outlined by the 
    Brooks, Pope and Marcolini (BPM) Model
    
    Assumptions:
        BPM models assumes a naca 0012 airfol  
        
    Source:  
        BPM Model:  Brooks, Thomas F., D. Stuart Pope, and Michael A.
        Marcolini. Airfoil self-noise and prediction. No. L-16528. 1989.
    
    Inputs:  
       R_c          - Reynolds number                        [-]
       alpha_star   - angle of attack blade section          [deg]
       delta_p      - boundary layer of pressure section     [m
       L            - length of blade section                [m]
       U            - velocity of blade section              [m/s]
       M            - Mach number                            [-]
       c            - airfoil section chord                  [m] 
       f            - frequency spectrum                     [Hz]
       Dbar_h       - high frequency directivity term        [-]
       r_e          - distance from noise source to observer [m] 
    
    Outputs 
       SPL_LBL_VS   - Sound pressure level of laminar boundary layer [dB]
       
    Properties Used:
        N/A   
    '''       
    
    St_prime              = f*delta_p/U # eqn 54 
    St_prime_1            = 0.001756*(R_c**0.3931)  # eqn 55 
    St_prime_1[R_c<1.3E5] = 0.18                    # eqn 55 
    St_prime_1[R_c>4E5]   = 0.28                    # eqn 55 
    
    St_prime_peak = St_prime_1*(10**(-0.04*alpha_star))
    
    G_1   = compute_G_1(St_prime/St_prime_peak)
    R_c_0 = compute_R_c_0(alpha_star)
    G_2   = compute_G_2(R_c/R_c_0)
    G_3   = compute_G_3(alpha_star)
    
    SPL_LBL_VS =  10*np.log10((delta_p*(M**5)*L*Dbar_h)/(r_e**2) ) + G_1  + G_2 + G_3 # eqn 53

    return  SPL_LBL_VS

def compute_G_1(e): 
    '''This computes the G_1 function using the BPM model
    
    Assumptions:
        BPM models assumes a naca 0012 airfol 
        Corrections made to match experimental results 
        
    Source:  
        BPM Model:  Brooks, Thomas F., D. Stuart Pope, and Michael A.
        Marcolini. Airfoil self-noise and prediction. No. L-16528. 1989.
    
    Inputs: 
        e     [-]
    
    Outputs 
        G_1   [-]
       
    Properties Used:
        N/A   
    '''         
    e     = e * 0.3  
    num_1 = 39.8 *0.5 
    num_2 = 98.409 *0.5  
    
    G_1           = -num_1*np.log10(e) - 11.2   # eqn 57
    G_1[e<1.64]   = -num_2*np.log10(e[e<1.64]) + 2  # eqn 57
    G_1[e<1.17]   = -5.076 + np.sqrt( 2.484 - 506.25*(np.log10(e[e<1.17]))**2)  # eqn 57
    G_1[e<0.8545] = num_2*np.log10(e[e<0.8545]) + 2      # eqn 57
    G_1[e<0.5974] = num_1*np.log10(e[e<0.5974]) - 11.2    # eqn 57
    
    return G_1

def compute_G_2(d):
    '''This computes the G_2 function using the BPM model
    
    Assumptions:
        BPM models assumes a naca 0012 airfol  
        
    Source:  
        BPM Model:  Brooks, Thomas F., D. Stuart Pope, and Michael A.
        Marcolini. Airfoil self-noise and prediction. No. L-16528. 1989.
    
    Inputs: 
        d     [-]
    
    Outputs 
        G_2   [-]
       
    Properties Used:
        N/A   
    '''     
 
    G_2           = -77.852*np.log10(d) + 15.328           # eqn 58 
    G_2[d<3.0889] = -65.188*np.log10(d[d<3.0889]) + 9.125  # eqn 58 
    G_2[d<1.7579] = -114.052*((np.log10(d[d<1.7579]))**2 ) # eqn 58
    G_2[d<0.5689] = 65.188*np.log10(d[d<0.5689]) + 9.125   # eqn 58
    G_2[d<0.3237] = 77.852*np.log10(d[d<0.3237]) + 15.328  # eqn 58    
    
    return G_2

def compute_G_3(alpha_star): 
    '''This computes the G_3 function using the BPM model
    
    Assumptions:
        BPM models assumes a naca 0012 airfol  
        
    Source:  
        BPM Model:  Brooks, Thomas F., D. Stuart Pope, and Michael A.
        Marcolini. Airfoil self-noise and prediction. No. L-16528. 1989.
    
    Inputs: 
        alpha_star    [deg]
    
    Outputs 
        G_3           [-]
       
    Properties Used:
        N/A   
    '''     
    G_3 = 171.04 - 3.03*alpha_star
    return G_3

def compute_R_c_0(alpha_star):
    '''This computes the R_c_0 function using the BPM model
    
    Assumptions:
        BPM models assumes a naca 0012 airfol  
        
    Source:  
        BPM Model:  Brooks, Thomas F., D. Stuart Pope, and Michael A.
        Marcolini. Airfoil self-noise and prediction. No. L-16528. 1989.
    
    Inputs: 
        alpha_star    [deg]
    
    Outputs 
        R_c_0         [-]
       
    Properties Used:
        N/A   
    '''     
    R_c_0               = 10**(0.215*alpha_star + 4.978) # eqn 59
    R_c_0[3<alpha_star] = 10**(0.120*alpha_star[3<alpha_star] + 5.263)  # eqn 59 
    return R_c_0 


# RCAIDE/Methods/Noise/Correlation_Buildup/Engine/mixed_noise_component.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
# Python package imports   
import numpy as np   

# ----------------------------------------------------------------------------------------------------------------------     
#  Mixed Noise Component
# ----------------------------------------------------------------------------------------------------------------------       
def mixed_noise_component(Velocity_primary,theta_m,sound_ambient,Velocity_secondary,
                          Velocity_aircraft,Area_primary,Area_secondary,DSPL_m,EX_m,Str_m,Velocity_mixed,XBPR):
    """This function calculates the noise contribution of the mixed jet component
    
    Assumptions:
        N/A

    Source:
        None

    Inputs:
        SPL_m               [dB]
        Velocity_primary    [m/s]
        theta_m             [rad]
        sound_ambient       [SPL]
        Velocity_secondary  [m/s]
        Velocity_aircraft   [m/s]
        Area_primary        [m^2]
        Area_secondary      [m^2]
        DSPL_m              [SPL]
        EX_m
        Str_m
        Velocity_mixed      [m/s]
        XBPR                   

    Outputs:
        SPL_m                [dB]

    Properties Used: 
        N/A 
    """ 

    #Calculation of the velocity exponent
    velocity_exponent = (Velocity_mixed/sound_ambient)**0.5*(0.6+(0.2/(0.2+Str_m) * \
        np.exp(-0.3*(theta_m+(Str_m/(1+Str_m))-2.7)**2)))

    #Calculation of the Source Strengh Function (FV)
    FV = ((Velocity_mixed-Velocity_aircraft)/sound_ambient)**velocity_exponent * \
        ((Velocity_mixed+Velocity_aircraft)/sound_ambient)**(1-velocity_exponent)

    #Determination of the noise model coefficients
    Z1 = -30*((1.8*theta_m/np.pi)-0.6)**2
    Z2 = -9 -4*((Velocity_primary-Velocity_secondary)/sound_ambient)-38*((1.8*theta_m/np.pi)-0.6)**3 + \
        30*(0.6-np.log10(1+Area_secondary/Area_primary))*(1.8*theta_m/np.pi - 0.6)
    Z3 = 1-0.4*((1.8*theta_m/np.pi)-0.6)**2
    Z4 = 0.44-0.5/np.exp(((4.5*theta_m/np.pi)-4)**2) + 0.2*Velocity_primary/sound_ambient - \
        0.7*Velocity_mixed/sound_ambient - 0.2*np.log10((1+Area_secondary)/Area_primary) + \
        0.05*(XBPR)*np.exp(-5*(theta_m-2.4)**2)
    Z5 = 34 + 81*theta_m/np.pi - 20*((1.8*theta_m/np.pi)-0.6)**3
    Z6 = 108 + 37.8*theta_m/np.pi + 5*Velocity_mixed*(Velocity_primary-Velocity_secondary)/(sound_ambient**2) - \
        np.exp(-5*(theta_m-1.8)**2) + 7*Velocity_mixed/sound_ambient*(1-0.4*(Velocity_primary/sound_ambient) * \
        np.exp(-0.7*np.abs(Str_m-0.8))) / np.exp(8*(theta_m-2.4)**2) + 0.8*(XBPR)*np.exp(theta_m-2.3-Velocity_mixed/sound_ambient) + \
        DSPL_m + EX_m

    #Determination of Sound Pressure Level for the mixed jet component
    SPL_m = (Z1*np.log10(FV)+Z2)*(np.log10(Str_m)-Z3*np.log10(FV)-Z4)**2 + Z5*np.log10(FV) + Z6

    return SPL_m 



# RCAIDE/Methods/Energy/Propulsors/Turbojet_Propulsor/size_core.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Library.Methods.Propulsors.Turbojet_Propulsor import compute_thrust

# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  size_core
# ---------------------------------------------------------------------------------------------------------------------- 
def size_core(turbojet,turbojet_conditions,conditions):
    """Sizes the core flow for the design condition.

    Assumptions:
    Perfect gas

    Source:
    https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/

    Inputs:
    conditions.freestream.speed_of_sound [m/s] (conditions is also passed to turbojet.compute(..))
    turbojet.inputs.
      bypass_ratio                       [-]
      total_temperature_reference        [K]
      total_pressure_reference           [Pa]
      number_of_engines                  [-]

    Outputs:
    turbojet.outputs.non_dimensional_thrust  [-]

    Properties Used:
    turbojet.
      reference_temperature              [K]
      reference_pressure                 [Pa]
      total_design                       [N] - Design thrust
    """             
    #unpack inputs
    a0                   = conditions.freestream.speed_of_sound
    throttle             = 1.0

    #unpack from turbojet 
    Tref                        = turbojet.reference_temperature
    Pref                        = turbojet.reference_pressure 

    total_temperature_reference = turbojet_conditions.total_temperature_reference  
    total_pressure_reference    = turbojet_conditions.total_pressure_reference 

    #compute nondimensional thrust
    turbojet_conditions.throttle = 1.0
    compute_thrust(turbojet,turbojet_conditions,conditions)

    #unpack results 
    Fsp                         = turbojet_conditions.non_dimensional_thrust

    #compute dimensional mass flow rates
    mdot_core                   = turbojet.design_thrust/(Fsp*a0*throttle)  
    mdhc                        = mdot_core/ (np.sqrt(Tref/total_temperature_reference)*(total_pressure_reference/Pref))

    #pack outputs
    turbojet.mass_flow_rate_design               = mdot_core
    turbojet.compressor_nondimensional_massflow  = mdhc

    return    



# RCAIDE/Compoments/Wings/Control_Surfaces/Slat.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from .Control_Surface import Control_Surface 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Slat
# ----------------------------------------------------------------------------------------------------------------------
class Slat(Control_Surface):
    """
    A class representing a slat control surface for high-lift generation during takeoff 
    and landing.

    Attributes
    ----------
    tag : str
        Unique identifier for the slat, defaults to 'slat'
        
    hinge_fraction : float
        Location of the hinge line as fraction of chord, defaults to 1.0
        (leading edge device)
        
    sign_duplicate : float
        Sign convention for duplicate slat deflection, defaults to 1.0
        (synchronized deflection for lift generation)

    Notes
    -----
    The slat is a leading edge high-lift device used to increase maximum lift coefficient 
    and stall angle at low speeds. It inherits basic control surface functionality from 
    the Control_Surface class and adds specific attributes for slat operation.
    
    **Definitions**

    'Hinge Fraction'
        The chordwise location of the slat hinge line, measured from the leading 
        edge as a fraction of local chord. Set to 1.0 for leading edge device
        
    'Sign Duplicate'
        Determines whether paired slats deflect in the same or opposite directions. 
        1.0 indicates synchronized deflection for lift generation

    See Also
    --------
    RCAIDE.Library.Components.Wings.Control_Surfaces.Control_Surface
        Base class providing common control surface functionality
    RCAIDE.Library.Components.Wings.Control_Surfaces.Flap
        Trailing edge high-lift device
    """ 

    def __defaults__(self):
        """
        Sets default values for the slat attributes.
        
        Notes
        -----
        See Control_Surface.__defaults__ for additional inherited attributes.
        """
        self.tag            = 'slat'
        self.hinge_fraction = 1.0
        self.sign_duplicate = 1.0
        
        pass 


# RCAIDE/Optimization/Packages/trmm/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from RCAIDE.Framework.Optimization.Packages      import trmm 
from RCAIDE.Framework.Optimization.Packages.trmm import TRMM_setup
from RCAIDE.Framework.Optimization.Packages.trmm import Trust_Region
from RCAIDE.Framework.Optimization.Packages.trmm import Trust_Region_Optimization


# RCAIDE/Methods/Stability/Center_of_Gravity/compute_vehicle_center_of_gravity.py 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# RCAIDE imports   
from RCAIDE.Library.Components import Component 
from RCAIDE.Library.Methods.Weights.mass_and_intertia_functions import *  
from .compute_component_centers_of_gravity import compute_component_centers_of_gravity

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Computer Aircraft Center of Gravity
# ----------------------------------------------------------------------------------------------------------------------   
def compute_vehicle_center_of_gravity(vehicle, update_CG=True): 
    ''' Computes the moment of intertia of aircraft 
    
    Source:
    Simplified Mass and Inertial Estimates for Aircraft with Components of Constant Density
    Moulton, B. C., and Hunsaker, D. F., “Simplified Mass and Inertial Estimates for Aircraft with Components 
    of Constant Density,” AIAA SCITECH 2023 Forum, January 2023, AIAA-2023-2432 DOI: 10.2514/
    6.2023-2432
    
    
    Assumtions:
    Assumes simplified shapes 
    
    Inputs:
    vehicle           - vehicle data structure           [m]
    
    Outputs:
    I                 - mass moment of inertia matrix    [kg-m^2]
    
    '''
    
    # compute compoment center of gravity     
    compute_component_centers_of_gravity(vehicle)
    
    # compute total aircraft center of grabity 
    total_moment = np.array([[0.0,0.0,0.0]])
    total_mass   = 0

    for key in vehicle.keys():
        item = vehicle[key]
        if isinstance(item,Component.Container):
            Moment, Mass  = sum_moment(item)  
            total_moment += Moment
            total_mass   += Mass         
    
    if update_CG:
        vehicle.mass_properties.center_of_gravity = total_moment/total_mass 
    
    CG =  total_moment/total_mass
   
    return CG, total_mass





# RCAIDE/Library/Plots/Geometry/plot_3d_vehicle_vlm_panelization.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
import RCAIDE
from RCAIDE.Library.Methods.Aerodynamics.Vortex_Lattice_Method  import generate_vortex_distribution
from RCAIDE.Library.Plots.Geometry.Common.contour_surface_slice import contour_surface_slice

import numpy as np  
import plotly.graph_objects as go 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------  
def plot_3d_vehicle_vlm_panelization(vehicle,
                                     alpha = 1.0,
                                     plot_axis   = False,
                                     save_figure = False,
                                     show_wing_control_points = True,
                                     save_filename = "VLM_Panelization",
                                     min_x_axis_limit            =  -5,
                                     max_x_axis_limit            =  40,
                                     min_y_axis_limit            =  -20,
                                     max_y_axis_limit            =  20,
                                     min_z_axis_limit            =  -20,
                                     max_z_axis_limit            =  20, 
                                     show_figure = True):
                                  
    """This plots vortex lattice panels created when Fidelity Zero  Aerodynamics 
    Routine is initialized

    Assumptions:
    None

    Source:
    None

    Inputs:
    vehicle.vortex_distribution

    Outputs: 
    Plots

    Properties Used:
    N/A	
    """

    # unpack vortex distribution
    try:
        VD = vehicle.vortex_distribution
    except:
        VL = RCAIDE.Framework.Analyses.Aerodynamics.Vortex_Lattice_Method()  
        VL.settings.number_of_spanwise_vortices  = 25
        VL.settings.number_of_chordwise_vortices = 5
        VL.settings.spanwise_cosine_spacing      = False
        VL.settings.model_fuselage               = False
        VL.settings.model_nacelle                = False
        VD = generate_vortex_distribution(vehicle,VL.settings) 

    camera        = dict(up=dict(x=0.5, y=0.5, z=1), center=dict(x=0, y=0, z=-.75), eye=dict(x=-1.5, y=-1.5, z=.8))
    plot_data     = []      

    # -------------------------------------------------------------------------
    # PLOT VORTEX LATTICE
    # -------------------------------------------------------------------------        
    n_cp      = VD.n_cp 
    color_map = 'greys'
    for i in range(n_cp):  
        X = np.array([[VD.XA1[i],VD.XA2[i]],[VD.XB1[i],VD.XB2[i]]])
        Y = np.array([[VD.YA1[i],VD.YA2[i]],[VD.YB1[i],VD.YB2[i]]])
        Z = np.array([[VD.ZA1[i],VD.ZA2[i]],[VD.ZB1[i],VD.ZB2[i]]])           
        
        values      = np.ones_like(X) 
        verts       = contour_surface_slice(X, Y, Z ,values,color_map)
        plot_data.append(verts)                 
  
  
    if  show_wing_control_points: 
        ctrl_pts = go.Scatter3d(x=VD.XC, y=VD.YC, z=VD.ZC,
                                    mode  = 'markers',
                                    marker= dict(size=6,color='red',opacity=0.8),
                                    line  = dict(color='red',width=2))
        plot_data.append(ctrl_pts)         
 
 
 
    fig = go.Figure(data=plot_data)
    fig.update_scenes(aspectmode   = 'auto',
                      xaxis_visible=plot_axis,
                      yaxis_visible=plot_axis,
                      zaxis_visible=plot_axis
                      )
    fig.update_layout( 
             width     = 1500,
             height    = 1500, 
             scene = dict(
                        xaxis = dict(backgroundcolor="grey", gridcolor="white", showbackground=plot_axis,
                                     zerolinecolor="white", range=[min_x_axis_limit,max_x_axis_limit]),
                        yaxis = dict(backgroundcolor="grey", gridcolor="white", showbackground=plot_axis, 
                                     zerolinecolor="white", range=[min_y_axis_limit,max_y_axis_limit]),
                        zaxis = dict(backgroundcolor="grey",gridcolor="white",showbackground=plot_axis,
                                     zerolinecolor="white", range=[min_z_axis_limit,max_z_axis_limit])),             
             scene_camera=camera) 
    fig.update_coloraxes(showscale=False)
    fig.update_traces(opacity = alpha)
    if save_figure:
        fig.write_image(save_filename + ".png")
        
    if show_figure:
        fig.write_html( save_filename + '.html', auto_open=True)
    return fig



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_systems_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import  RCAIDE 
from RCAIDE.Framework.Core    import Units, Data 

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# Systems Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_systems_weight(vehicle):
    """ Calculate the system weight of the aircraft including:
        -  flight control system
        - apu weight
        - hydraulics and pneumatics weight
        - intstrumentation weight
        - avionics weight
        - electrical system weight
        - air-condtioning weight
        - furnishing weight
        - anti-ice weight

    Assumptions:
        1) No variable sweep, change VARSWP to 1 is variable sweep in aicraft
        2) Hydraulic pressure is 3000 psf (HYDR)
        3) 1 fuselage named fuselage. Function could be modified to allow multiple fuselages by
           relaxing this assumption.

    Source:
        The Flight Optimization System Weight Estimation Method

   Inputs:
        vehicle - data dictionary with vehicle properties                   [dimensionless]
            -.networks: data dictionary containing all propulsion properties
                -.number_of_engines: number of engines
                -.sealevel_static_thrust: thrust at sea level               [N]
            -.fuselages['fuselage'].lengths.total: fuselage total length    [meters]
            -.fuselages['fuselage'].width: fuselage width                   [meters]
            -.fuselages['fuselage'].heights.maximum: fuselage maximum height[meters]
            -.mass_properties.max_takeoff: MTOW                             [kilograms]
            -.design_mach_number: design mach number for cruise flight
            -.design_range: design range of aircraft                        [nmi]
            -.passengers: number of passengers in aircraft
            -.wings['main_wing']: data dictionary with main wing properties
                -.sweeps.quarter_chord: quarter chord sweep                 [deg]
                -.areas.reference: wing surface area                        [m^2]
                -.spans.projected: projected span of wing                   [m]
                -.flap_ratio: flap surface area over wing surface area

   Outputs:
       output - a data dictionary with fields:
           W_flight_controls - weight of the flight control system                                [kilograms]
           W_apu - weight of the apu                                                       [kilograms]
           W_hyd_pnu - weight of the hydraulics and pneumatics                             [kilograms]
           W_instruments - weight of the instruments and navigational equipment            [kilograms]
           W_avionics - weight of the avionics                                             [kilograms]
           W_electrical - weight of the electrical items                                         [kilograms]
           W_ac - weight of the air conditioning and anti-ice system                       [kilograms]
           W_furnish - weight of the furnishings in the fuselage                           [kilograms]
           W_anti_ice - weight of anti-ice system                                          [kilograms]

    Properties Used:
        N/A
    """ 
    NENG = 0
    FNEW = 0
    FNEF = 0 
    for network in  vehicle.networks:
        for propulsor in network.propulsors:
            if isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbojet):
                NENG += 1 
                FNEF += 1
                if propulsor.wing_mounted: 
                    FNEW += 1   
                if 'nacelle' in propulsor:
                    nacelle =  propulsor.nacelle 
                    FNAC    = nacelle.diameter / Units.ft
                else:
                    FNAC    = 0                     
            
    VMAX     = vehicle.flight_envelope.design_mach_number
    SFLAP    = 0
    ref_wing = None 
    for wing in  vehicle.wings:
        if isinstance(wing, RCAIDE.Library.Components.Wings.Main_Wing):
            SFLAP  += wing.areas.reference * wing.flap_ratio / Units.ft ** 2
            ref_wing  =  wing
    
    S = 0
    if ref_wing == None:
        for wing in  vehicle.wings:
            if S < wing.areas.reference:
                ref_wing = wing
                
    DG    = vehicle.mass_properties.max_takeoff / Units.lbs
    WSC   = 1.1 * VMAX ** 0.52 * SFLAP ** 0.6 * DG ** 0.32  # surface controls weight
    
    XL = 0
    WF = 0
    L_fus = 0
    for fuselage in vehicle.fuselages:
        if L_fus < fuselage.lengths.total:
            ref_fuselage = fuselage 
            XL  = fuselage.lengths.total / Units.ft
            WF  = fuselage.width / Units.ft
    FPAREA      = XL * WF
    NPASS       = vehicle.passengers
    WAPU        = 54 * FPAREA ** 0.3 + 5.4 * NPASS ** 0.9  # apu weight

    if vehicle.passengers >= 150:
        NFLCR = 3  # number of flight crew
    else:
        NFLCR = 2 
    WIN     = 0.48 * FPAREA ** 0.57 * VMAX ** 0.5 * (10 + 2.5 * NFLCR + FNEW + 1.5 * FNEF)  # instrumentation weight

    SW      = vehicle.reference_area / Units.ft ** 2
    HYDR    = 3000  # Hydraulic system pressure
    VARSWP  = 0
    WHYD    = 0.57 * (FPAREA + 0.27 * SW) * (1 + 0.03 * FNEW + 0.05 * FNEF) * (3000 / HYDR) ** 0.35 * \
            (1 + 0.04 * VARSWP) * VMAX ** 0.33  # hydraulic and pneumatic system weight

    NFUSE   = len(vehicle.fuselages)
    WELEC   = 92. * XL ** 0.4 * WF ** 0.14 * NFUSE ** 0.27 * NENG ** 0.69 * \
            (1. + 0.044 * NFLCR + 0.0015 * NPASS)  # electrical system weight

    DESRNG  = vehicle.flight_envelope.design_range / Units.nmi
    WAVONC  = 15.8 * DESRNG ** 0.1 * NFLCR ** 0.7 * FPAREA ** 0.43  # avionics weight

    XLP     = 0.8 * XL
    DF      = ref_fuselage.heights.maximum / Units.ft # D stands for depth
    WFURN   = 127 * NFLCR + 112 * vehicle.NPF + 78 * vehicle.NPB + 44 * vehicle.NPT \
                + 2.6 * XLP * (WF + DF) * NFUSE  # furnishing weight

    WAC     = (3.2 * (FPAREA * DF) ** 0.6 + 9 * NPASS ** 0.83) * VMAX + 0.075 * WAVONC  # ac weight

    WAI     = ref_wing.spans.projected / Units.ft * 1. / np.cos(ref_wing.sweeps.quarter_chord) + 3.8 * FNAC * NENG + 1.5 * WF  # anti-ice weight

    output                      = Data()
    output.W_flight_control    = WSC * Units.lbs
    output.W_apu               = WAPU * Units.lbs
    output.W_hyd_pnu           = WHYD * Units.lbs
    output.W_instruments       = WIN * Units.lbs
    output.W_avionics          = WAVONC * Units.lbs
    output.W_electrical        = WELEC * Units.lbs
    output.W_ac                = WAC * Units.lbs
    output.W_furnish           = WFURN * Units.lbs
    output.W_anti_ice          = WAI * Units.lbs
    output.W_systems           = WSC + WAPU + WIN + WHYD + WELEC + WAVONC + WFURN + WAC + WAI
    return output



## @ingroup Attributes-Solids

# Epoxy.py
#
# Created: Jul, 2017, J. Smart
# ModifiedL Apr, 2018, J. Smart

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

from .Solid import Solid
from RCAIDE.Framework.Core import Units

#-------------------------------------------------------------------------------
# Hardened Epoxy Resin Solid Class
#-------------------------------------------------------------------------------

## @ingroup Attributes-Solids
class Epoxy(Solid):
    """ Physical Constants Specific to Epoxy Resin
    
    Assumptions:
    None
    
    Source:
    MatWeb (Median of Mfg. Reported Values)
    
    Inputs:
    N/A
    
    Outputs:
    N/A
    
    Properties Used:
    None
    """

    def __defaults__(self):
        """Sets material properties at instantiation.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        None
        """

        self.ultimate_tensile_strength  = 0.0       * Units.Pa
        self.ultimate_shear_strength    = 0.0       * Units.Pa
        self.ultimate_bearing_strength  = 0.0       * Units.Pa
        self.yield_tensile_strength     = 0.0       * Units.Pa
        self.yield_shear_strength       = 0.0       * Units.Pa
        self.yield_bearing_strength     = 0.0       * Units.Pa
        self.minimum_gage_thickness     = 250e-6    * Units.m
        self.density                    = 1800.     * Units['kg/(m**3)']




#RCAIDE/Library/Methods/Thermal_Management/Batteries/Liquid_Cooled_Wavy_Channel/wavy_channel_geometry_setup.py
#
# Created: Apr 2024, S. Shekar 2024

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core        import Data

# ----------------------------------------------------------------------------------------------------------------------  
#  Wavy Channel Geometry Setup 
# ----------------------------------------------------------------------------------------------------------------------   

def wavy_channel_geometry_setup(HAS,battery): 
    """ Defines a dummy vehicle for wavy channel geometry optimization.
          
          Inputs:  
             HAS   - HAS data structure             [None] 
             battery - Battery data structure
         Outputs:  
             configs - configuration used in optimization    [None]
              
          Assumptions: 
             N/A 
        
          Source:
             None
    """     
    vehicle                                                   = RCAIDE.Vehicle()  
    net                                                       = RCAIDE.Framework.Networks.Electric()
    bus                                                       = RCAIDE.Library.Components.Energy.Distributors.Electrical_Bus()
    bus.battery_modules.append(battery)
    coolant_line                                                          = RCAIDE.Library.Components.Energy.Distributors.Coolant_Line(bus)     
    coolant_line.battery_modules[battery.tag].thermal_management_system   = Data() 

    coolant_line.battery_modules[battery.tag].thermal_management_system.heat_acquisition_system = HAS 

    net.busses.append(bus)
    net.coolant_lines.append(coolant_line) 
    vehicle.append_energy_network(net) 
    
    configs                                                     = RCAIDE.Library.Components.Configs.Config.Container()
    base_config                                                 = RCAIDE.Library.Components.Configs.Config(vehicle)  
    config                                                      = RCAIDE.Library.Components.Configs.Config(base_config)
    config.tag                                                  = 'optimized'  
    configs.append(config)         
    return configs 


# RCAIDE/Library/Compoments/Energy/Propulsion/Converters/Rotor.py
# 
# 
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

 # RCAIDE imports 
from RCAIDE.Framework.Core                              import Data , Units, Container
from RCAIDE.Library.Components                          import Component 
from RCAIDE.Framework.Analyses.Propulsion               import Momentum_Theory_Wake 
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.append_rotor_conditions import  append_rotor_conditions

# package imports
import numpy as np
import scipy as sp

# ---------------------------------------------------------------------------------------------------------------------- 
#  Generalized Rotor Class
# ---------------------------------------------------------------------------------------------------------------------- 
class Rotor(Component):
    """
    A generalized rotor component model serving as the base class for various rotary propulsion devices.

    Attributes
    ----------
    tag : str
        Identifier for the rotor. Default is 'rotor'.
        
    number_of_blades : float
        Number of rotor blades. Default is 0.0.
        
    tip_radius : float
        Outer radius of the rotor [m]. Default is 0.0.
        
    hub_radius : float
        Inner radius of the rotor at hub [m]. Default is 0.0.
        
    twist_distribution : float
        Blade twist angle distribution [rad]. Default is 0.0.
        
    sweep_distribution : float
        Quarter chord sweep distribution [m]. Default is 0.0.
        
    chord_distribution : float
        Blade chord length distribution [m]. Default is 0.0.
        
    thickness_to_chord : float
        Ratio of blade thickness to chord. Default is 0.0.
        
    max_thickness_distribution : float
        Maximum thickness distribution along blade [m]. Default is 0.0.
        
    radius_distribution : ndarray
        Radial stations along blade [m]. Default is None.
        
    mid_chord_alignment : float
        Mid-chord line alignment [m]. Default is 0.0.
        
    blade_solidity : float
        Ratio of total blade area to rotor disk area. Default is 0.0.
        
    flap_angle : float
        Blade flapping angle [rad]. Default is 0.0.
        
    number_azimuthal_stations : int
        Number of azimuthal calculation points. Default is 16.
        
    vtk_airfoil_points : int
        Number of points for VTK airfoil visualization. Default is 40.
        
    Airfoils : Airfoil_Container
        Container for blade airfoil definitions. Default is empty container.
        
    airfoil_polar_stations : ndarray
        Radial stations for airfoil polars. Default is None.
        
    induced_power_factor : float
        Factor accounting for non-ideal induced power. Default is 1.48.
        
    profile_drag_coefficient : float
        Mean blade profile drag coefficient. Default is 0.03.
        
    clockwise_rotation : bool
        Direction of rotation. Default is True.
        
    phase_offset_angle : float
        Initial blade phase angle [rad]. Default is 0.0.
        
    orientation_euler_angles : list
        Angles defining rotor orientation [rad]. Default is [0.,0.,0.].
        
    pitch_command : float
        Commanded blade pitch angle [rad]. Default is 0.0.
        
    ducted : bool
        Flag for ducted rotor configuration. Default is False.
        
    sol_tolerance : float
        Solution convergence tolerance. Default is 1e-8.
        
    use_2d_analysis : bool
        Flag for 2D aerodynamic analysis. Default is False.
        
    nonuniform_freestream : bool
        Flag for nonuniform inflow conditions. Default is False.

    Notes
    -----
    The Rotor class provides a comprehensive framework for modeling rotary
    propulsion devices including:

    * Geometric definition of rotor and blades
    * Aerodynamic performance calculation
    * Wake modeling capabilities
    * Blade element analysis
    * Performance optimization
    * Acoustic analysis

    **Major Assumptions**

    * Rigid blade structure
    * Quasi-steady aerodynamics
    * Small angle approximations for flapping
    * Linear blade twist and taper
    * Uniform inflow (unless nonuniform_freestream is True)

    **Theory**

    The rotor model combines:
    * Blade Element Momentum Theory
    * Prescribed/Free Wake Analysis
    * Acoustic Propagation Models
    * Performance Optimization Methods

    **Definitions**

    'Blade Solidity'
        Ratio of total blade area to rotor disk area
    'Induced Power Factor'
        Correction for non-ideal induced power effects
    'Phase Offset'
        Initial azimuthal position of reference blade

    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Converters.Propeller
    RCAIDE.Library.Components.Propulsors.Converters.Lift_Rotor
    RCAIDE.Library.Components.Propulsors.Converters.Prop_Rotor
    """
    def __defaults__(self):
        """This sets the default values for the component to function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        None
        """

        self.tag                               = 'rotor' 
        
        # geometry properties 
        self.number_of_blades                  = 0.0
        self.tip_radius                        = 0.0
        self.hub_radius                        = 0.0
        self.twist_distribution                = 0.0
        self.sweep_distribution                = 0.0         # quarter chord offset from quarter chord of root airfoil
        self.chord_distribution                = 0.0
        self.thickness_to_chord                = 0.0
        self.max_thickness_distribution        = 0.0
        self.radius_distribution               = None
        self.mid_chord_alignment               = 0.0
        self.blade_solidity                    = 0.0 
        self.flap_angle                        = 0.0
        self.number_azimuthal_stations         = 16 
        self.vtk_airfoil_points                = 40        
        self.airfoils                          = Airfoil_Container()
        self.airfoil_polar_stations            = None 
        
        # design flight conditions 
        self.cruise                            = Data() 
        self.cruise.design_power               = None
        self.cruise.design_thrust              = None
        self.cruise.design_power_coefficient   = 0.01 

        # operating conditions 
        self.induced_power_factor              = 1.48        # accounts for interference effects
        self.profile_drag_coefficient          = .03
        self.clockwise_rotation                = True
        self.phase_offset_angle                = 0.0
        self.orientation_euler_angles          = [0.,0.,0.]  # vector of angles defining default orientation of rotor
        self.pitch_command                     = 0.0
        self.ducted                            = False
        self.sol_tolerance                     = 1e-8 
        self.use_2d_analysis                   = False       # True if rotor is at an angle relative to freestream or nonuniform freestream
        self.nonuniform_freestream             = False   
        self.axial_velocities_2d               = None        # user input for additional velocity influences at the rotor
        self.tangential_velocities_2d          = None        # user input for additional velocity influences at the rotor
        self.radial_velocities_2d              = None        # user input for additional velocity influences at the rotor 
        self.start_angle                       = 0.0         # angle of first blade from vertical
        self.start_angle_idx                   = 0           # azimuthal index at which the blade is started 
        self.variable_pitch                    = False
        self.electric_propulsion_fraction      = 1.0

        # Initialize the default wake set to Fidelity Zero 
        self.Wake                      = Momentum_Theory_Wake() 
        
        # blade optimization parameters     
        self.optimization_parameters                                    = Data() 
        self.optimization_parameters.tip_mach_range                     = [0.3,0.7] 
        self.optimization_parameters.multiobjective_aeroacoustic_weight = 1.0
        self.optimization_parameters.multiobjective_performance_weight  = 1.0
        self.optimization_parameters.multiobjective_acoustic_weight     = 1.0
        self.optimization_parameters.noise_evaluation_angle             = 135 * Units.degrees 
        self.optimization_parameters.tolerance                          = 1E-4
        self.optimization_parameters.ideal_SPL_dBA                      = 30
        self.optimization_parameters.ideal_efficiency                   = 1.0     
        self.optimization_parameters.ideal_figure_of_merit              = 1.0

    def append_operating_conditions(rotor,segment,propulsor): 
        energy_conditions       = segment.state.conditions.energy[propulsor.tag]
        noise_conditions        = segment.state.conditions.noise[propulsor.tag]
        append_rotor_conditions(rotor,segment,energy_conditions,noise_conditions)
        return        
         
    def append_airfoil(self,airfoil):
        """ Adds an airfoil to the segment

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """  
        # assert database type
        if not isinstance(airfoil,Data):
            raise Exception('input component must be of type Data()')


        # See if the component exists, if it does modify the name
        keys = self.keys()
        if airfoil.tag in keys:
            string_of_keys = "".join(self.keys())
            n_comps = string_of_keys.count(airfoil.tag)
            airfoil.tag = airfoil.tag + str(n_comps+1)    
            
        # store data
        self.airfoils.append(airfoil)
    
    def vec_to_vel(self):
        """
        Rotates from the rotor's vehicle frame to the rotor's velocity frame.

        Returns
        -------
        rot_mat : ndarray
            3x3 rotation matrix transforming from vehicle frame to velocity frame.

        Notes
        -----
        This method creates a rotation matrix for transforming coordinates between
        two reference frames of the rotor. When rotor is axially aligned with the 
        vehicle body.

        Velocity frame:

        * X-axis points out the nose
        * Z-axis points towards the ground
        * Y-axis points out the right wing

        Vehicle frame:

        * X-axis points towards the tail
        * Z-axis points towards the ceiling
        * Y-axis points out the right wing

        **Major Assumptions**

        * The rotor's default orientation is aligned with the vehicle body
        * Right-handed coordinate system is used
        * Small angle approximations are not used
        * Rotation sequence is fixed
        """

        rot_mat = sp.spatial.transform.Rotation.from_rotvec([0,np.pi,0]).as_matrix()

        return rot_mat
    

    def body_to_prop_vel(self, commanded_thrust_vector):
        """
        Rotates from the system's body frame to the rotor's velocity frame.

        Parameters
        ----------
        commanded_thrust_vector : ndarray
            Vector of commanded thrust angles [rad] for each time step.

        Returns
        -------
        rot_mat : ndarray
            3x3 rotation matrix transforming from body frame to rotor velocity frame.
        rots : ndarray
            Array of rotation vectors including commanded thrust angles and orientation.

        Notes
        -----
        This method performs a sequence of rotations to transform coordinates from
        the vehicle body frame to the rotor's velocity frame. The transformation
        sequence is:

        1. Body to vehicle frame (π rotation about Y-axis)
        2. Vehicle to rotor vehicle frame (includes thrust vector and orientation)
        3. Rotor vehicle to rotor velocity frame

        **Theory**
        The complete transformation is computed as:
        .. math::
            R_{total} = (R_{body2vehicle} R_{vehicle2rotor}) R_{rotor2vel}

        Where:

        * R_{body2vehicle} is a π rotation about Y-axis
        * R_{vehicle2rotor} includes orientation_euler_angles and thrust command
        * R_{rotor2vel} is from vec_to_vel()

        Velocity frame:

        * X-axis points out the nose
        * Z-axis points towards the ground
        * Y-axis points out the right wing

        Vehicle frame:

        * X-axis points towards the tail
        * Z-axis points towards the ceiling
        * Y-axis points out the right wing

        **Major Assumptions**

        * The rotor's default orientation is defined by orientation_euler_angles
        * Right-handed coordinate system is used
        * Thrust vector rotation is applied about the Y-axis
        * Matrix multiplication order preserves proper transformation sequence
        * Euler angle sequence is fixed
        """

        # Go from velocity to vehicle frame
        body_2_vehicle = sp.spatial.transform.Rotation.from_rotvec([0,np.pi,0]).as_matrix()

        # Go from vehicle frame to propeller vehicle frame: rot 1 including the extra body rotation
        cpts       = len(np.atleast_1d(commanded_thrust_vector))
        rots       = np.array(self.orientation_euler_angles) * 1.
        rots       = np.repeat(rots[None,:], cpts, axis=0) 
        rots[:,1] += commanded_thrust_vector[:,0] 
        
        vehicle_2_prop_vec = sp.spatial.transform.Rotation.from_rotvec(rots).as_matrix()

        # GO from the propeller vehicle frame to the propeller velocity frame: rot 2
        prop_vec_2_prop_vel = self.vec_to_vel()

        # Do all the matrix multiplies
        rot1    = np.matmul(body_2_vehicle,vehicle_2_prop_vec)
        rot_mat = np.matmul(rot1,prop_vec_2_prop_vel)
 
        return rot_mat , rots


    def prop_vel_to_body(self, commanded_thrust_vector):
        """
        Rotates from the rotor's velocity frame to the system's body frame.

        Parameters
        ----------
        commanded_thrust_vector : ndarray
            Vector of commanded thrust angles [rad] for each time step.

        Returns
        -------
        rot_mat : ndarray
            3x3 rotation matrix transforming from rotor velocity frame to body frame.
        rots : ndarray
            Array of rotation vectors including commanded thrust angles and orientation.

        Notes
        -----
        This method performs the inverse transformation sequence of body_to_prop_vel.
        The transformation sequence is:

        1. Rotor velocity to rotor vehicle frame
        2. Rotor vehicle to vehicle frame (includes thrust vector and orientation)
        3. Vehicle to body frame (π rotation about Y-axis)

        **Theory**
        The complete transformation is computed as:
        .. math::
            R_{total} = (R_{body2propvel})^{-1}

        Velocity frame:

        * X-axis points out the nose
        * Z-axis points towards the ground
        * Y-axis points out the right wing

        Vehicle frame:

        * X-axis points towards the tail
        * Z-axis points towards the ceiling
        * Y-axis points out the right wing

        **Major Assumptions**

        * The rotor's default orientation is defined by orientation_euler_angles
        * Right-handed coordinate system is used
        * Thrust vector rotation is applied about the Y-axis
        * Rotation matrices are orthogonal (inverse = transpose)
        * Euler angle sequence is fixed
        """

        body2propvel,rots = self.body_to_prop_vel(commanded_thrust_vector)

        r = sp.spatial.transform.Rotation.from_matrix(body2propvel)
        r = r.inv()
        rot_mat = r.as_matrix()

        return rot_mat, rots
    
    def vec_to_prop_body(self,commanded_thrust_vector):
        rot_mat, rots =  self.prop_vel_to_body(commanded_thrust_vector) 
        return rot_mat, rots

 
class Airfoil_Container(Container):
    """ Container for rotor airfoil  
    
    Assumptions:
    None

    Source:
    N/A

    Inputs:
    None

    Outputs:
    None

    Properties Used:
    N/A
    """     

    def get_children(self):
        """ Returns the components that can go inside
        
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:
        None
    
        Outputs:
        None
    
        Properties Used:
        N/A
        """       
        
        return []



# RCAIDE/Library/Missions/Common/Update/moments.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Framework.Core  import  orientation_product  

# ----------------------------------------------------------------------------------------------------------------------
#  Update Moments
# ----------------------------------------------------------------------------------------------------------------------
def moments(segment): 
    """ Updates the total resultant moment on the vehicle 
        
        Assumptions:
        N/A
        
        Inputs:
            segment.state.conditions.:
                frames.wind.force_vector          [N] 
                frames.body.thrust_force_vector        [N]
                frames.inertial.gravity_force_vector   [N]
        Outputs:
            segment.conditions
                frames.inertial.total_force_vector     [N]
 
      
        Properties Used:
        N/A
                    
    """    

    # unpack
    conditions                = segment.state.conditions  
    wind_moment_vector        = conditions.frames.wind.moment_vector
    body_thrust_moment_vector = conditions.frames.body.thrust_moment_vector
    
    # unpack transformation matrices
    T_wind2inertial = conditions.frames.wind.transform_to_inertial 
    T_body2inertial = conditions.frames.body.transform_to_inertial 

    # to inertial frame 
    M_t = orientation_product(T_body2inertial,body_thrust_moment_vector) 
    M_w = orientation_product(T_wind2inertial, wind_moment_vector) 
    
    M = M_t + M_w
    
    # pack
    conditions.frames.inertial.total_moment_vector[:,:] = M[:,:]

    return



## @ingroup Methods-Performance
# estimate_landing_field_length.py
#
# Created:  Jun 2014, T. Orra, C. Ilario, Celso, 
# Modified: Apr 2015, M. Vegh 
#           Jan 2016, E. Botero 
#           Mar 2020, M. Clarke
#           Jul 2020, E. Botero 

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
import  RCAIDE
from   RCAIDE.Framework.Core import Data, Units
from   RCAIDE.Library.Methods.Aerodynamics.Common.Lift.compute_max_lift_coeff import compute_max_lift_coeff

import numpy as np

# ----------------------------------------------------------------------
#  Compute field length required for landing
# ----------------------------------------------------------------------

## @ingroup Methods-Performance
def estimate_landing_field_length(vehicle,analyses, altitude=0, delta_isa=0):
    """ Computes the landing field length for a given vehicle configuration in a given airport.

    Assumptions:
    See source
    Two wheel trucks (code needed for four wheel trucks also included)

    Source:
    Torenbeek, E., "Advanced Aircraft Design", 2013 (equation 9.25)

    Inputs:
    airport.
      atmosphere                           [RCAIDE data type]
      altitude                             [m]
      delta_isa                            [K]
    vehicle.
      mass_properties.landing              [kg]
      reference_area                       [m^2]
      maximum_lift_coefficient (optional)  [Unitless]

    Outputs:
    landing_field_length                   [m]

    Properties Used:
    N/A
    """       
   
    # ==============================================
    # Unpack
    # ============================================== 
    altitude        = altitude * Units.ft
    delta_isa       = delta_isa
    weight          = vehicle.mass_properties.landing
    reference_area  = vehicle.reference_area
    try:
        Vref_VS_ratio = vehicle.Vref_VS_ratio
    except:
        Vref_VS_ratio = 1.23
        
    # ==============================================
    # Computing atmospheric conditions
    # ==============================================
    atmo            = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_values     = atmo.compute_values(altitude,delta_isa)
    
    p                 = atmo_values.pressure
    T                 = atmo_values.temperature
    rho               = atmo_values.density
    a                 = atmo_values.speed_of_sound
    mu                = atmo_values.dynamic_viscosity
    sea_level_gravity = atmo.planet.sea_level_gravity
   
    # ==============================================
    # Determining vehicle maximum lift coefficient
    # ==============================================
    # Condition to CLmax calculation: 90KTAS @ airport
    state = Data()
    state.conditions =  RCAIDE.Framework.Mission.Common.Results() 
    state.conditions.freestream = Data()
    state.conditions.freestream.density           = rho
    state.conditions.freestream.velocity          = 90. * Units.knots
    state.conditions.freestream.dynamic_viscosity = mu
    
    settings = analyses.aerodynamics.settings

    maximum_lift_coefficient, induced_drag_high_lift = compute_max_lift_coeff(state,settings,vehicle)

    # ==============================================
    # Computing speeds (Vs, Vref)
    # ==============================================
    stall_speed  = (2 * weight * sea_level_gravity / (rho * reference_area * maximum_lift_coefficient)) ** 0.5
    Vref         = stall_speed * Vref_VS_ratio
    
    # ========================================================================================
    # Computing landing distance, according to Torenbeek equation
    #     Landing Field Length = k1 + k2 * Vref**2
    # ========================================================================================

    # Defining landing distance equation coefficients 
    landing_constants    = np.zeros(3)
    landing_constants[0] = 250.
    landing_constants[1] =   0.
    landing_constants[2] =  2.485  / sea_level_gravity  # Two-wheels truck : [ (1.56 / 0.40 + 1.07) / (2*sea_level_gravity) ]
    
    # Calculating landing field length
    landing_field_length = 0.
    for idx,constant in enumerate(landing_constants):
        landing_field_length += constant * Vref**idx
    
    # return
    return landing_field_length



# RCAIDE/Framework/Analyses/Mission/Segments/Descent/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Constant_Speed_Constant_Angle       import Constant_Speed_Constant_Angle
from .Constant_Speed_Constant_Angle_Noise import Constant_Speed_Constant_Angle_Noise
from .Constant_Speed_Constant_Rate        import Constant_Speed_Constant_Rate
from .Linear_Mach_Constant_Rate           import Linear_Mach_Constant_Rate
from .Linear_Speed_Constant_Rate          import Linear_Speed_Constant_Rate
from .Constant_EAS_Constant_Rate          import Constant_EAS_Constant_Rate
from .Constant_CAS_Constant_Rate          import Constant_CAS_Constant_Rate


# RCAIDE/Library/Missions/Common/Pre_Process/stability.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports  
import  RCAIDE 
from RCAIDE.Library.Methods.Geometry.Planform  import wing_segmented_planform, wing_planform
# ----------------------------------------------------------------------------------------------------------------------
#  stability
# ----------------------------------------------------------------------------------------------------------------------  
def stability(mission):
    """ Runs stability model and build surrogate
    
        Assumptions:
            N/A
        
        Inputs:
            None
            
        Outputs:
            None             
    """
    last_tag = None
    for tag,segment in mission.segments.items(): 
                
        if segment.analyses.stability !=  None: 
            # ensure all properties of wing are computed before drag calculations  
            vehicle =  segment.analyses.stability.vehicle
            for wing in  vehicle.wings: 
                if len(wing.segments) > 1: 
                    wing_segmented_planform(wing)
                else:
                    wing_planform(wing)
                    
            if  (last_tag!=  None) and  ('compute' in mission.segments[last_tag].analyses.stability.process.keys()): 
                segment.analyses.stability.process.compute.lift.inviscid_wings = mission.segments[last_tag].analyses.stability.process.compute.lift.inviscid_wings
                segment.analyses.stability.surrogates       = mission.segments[last_tag].analyses.stability.surrogates 
                segment.analyses.stability.reference_values = mission.segments[last_tag].analyses.stability.reference_values   
            else: # use aerodynamic results that have been previously processed 
                if (type(segment.analyses.aerodynamics) == RCAIDE.Framework.Analyses.Aerodynamics.Vortex_Lattice_Method) or\
                (type(segment.analyses.aerodynamics) == RCAIDE.Framework.Analyses.Aerodynamics.Athena_Vortex_Lattice) :
                    segment.analyses.stability.process.compute.lift.inviscid_wings = segment.analyses.aerodynamics.process.compute.lift.inviscid_wings 
                    segment.analyses.stability.surrogates       = segment.analyses.aerodynamics.surrogates 
                    segment.analyses.stability.reference_values = segment.analyses.aerodynamics.reference_values 
                    last_tag = tag                 
                else: # run new simulation 
                    stab = segment.analyses.stability
                    stab.initialize() 
                    last_tag = tag 
    return 


# RCAIDE/Library/Plots/Energy/plot_electric_propulsor_efficiencies.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style 
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_electric_propulsor_efficiencies(results,
                                  save_figure = False,
                                  show_legend=True,
                                  save_filename = "Electric_Efficiencies",
                                  file_type = ".png",
                                  width = 11, height = 7):
    """This plots the electric driven network propeller efficiencies 

    Assumptions:
    None

    Source:
    None

    Inputs:
    results.segments.conditions.propulsion. 
         etap
         etam
         fom
        
    Outputs: 
    Plots

    Properties Used:
    N/A	
    """	   
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters) 
    
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     
    
    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)   
    axis_1 = plt.subplot(2,2,1)
    axis_2 = plt.subplot(2,2,2) 
    axis_3 = plt.subplot(2,2,3)
    

    for network in results.segments[0].analyses.energy.vehicle.networks:  
        for p_i, propulsor in enumerate(network.propulsors):
            if (p_i == 0) or (network.identical_propulsors == False): 
                for i in range(len(results.segments)):  
                    if 'rotor' in propulsor: 
                        thrustor =  propulsor.rotor
                        axis_1.set_ylabel(r'$\eta_{rotor}$')
                    elif 'ducted_fan' in propulsor:
                        thrustor =  propulsor.ducted_fan
                        axis_1.set_ylabel(r'$\eta_{ducted fan}$')
                    motor =  propulsor.motor
                      
                    bus_results  = results.segments[i].conditions.energy 
                    time         = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min      
                    effp         = bus_results[propulsor.tag][thrustor.tag].efficiency[:,0]
                    fom          = bus_results[propulsor.tag][thrustor.tag].figure_of_merit[:,0]
                    effm         = bus_results[propulsor.tag][motor.tag].efficiency[:,0]  
                    
                    if p_i == 0 and i ==0:              
                        axis_1.plot(time, effp, color = line_colors[i], marker = ps.markers[p_i], markersize= ps.marker_size, linewidth = ps.line_width, label = thrustor.tag)
                    else:
                        axis_1.plot(time, effp, color = line_colors[i], marker = ps.markers[p_i], markersize= ps.marker_size, linewidth = ps.line_width) 
                    axis_1.set_ylim([0,1.1])
                    set_axes(axis_1)         
                     
                    axis_2.plot(time, fom, color = line_colors[i], marker = ps.markers[p_i], markersize= ps.marker_size , linewidth = ps.line_width)
                    axis_2.set_xlabel('Time (mins)')
                    axis_2.set_ylabel(r'FoM')
                    axis_2.set_ylim([0,1.1])
                    set_axes(axis_2) 
             
                    axis_3.plot(time, effm, color = line_colors[i], marker = ps.markers[p_i], markersize= ps.marker_size, linewidth = ps.line_width)
                    axis_3.set_xlabel('Time (mins)')
                    axis_3.set_ylabel(r'$\eta_{motor}$')
                    axis_3.set_ylim([0,1.1])
                    set_axes(axis_3)
           
    if show_legend:     
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4)  
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8) 
    
    # set title of plot 
    title_text  =  'Electronic Network Efficiencies' 
    fig.suptitle(title_text)
    if save_figure:
        plt.savefig(save_filename + file_type) 
     
    return fig  


# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/parasite_drag_nacelle.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
  
from RCAIDE.Framework.Core                    import Data  
from RCAIDE.Library.Methods.Utilities         import Cubic_Spline_Blender   
from RCAIDE.Library.Methods.Aerodynamics.Common.Drag.compressible_turbulent_flat_plate import compressible_turbulent_flat_plate

# package imports
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Supersonic Parasite Drag Nacekke 
# ---------------------------------------------------------------------------------------------------------------------- 
def parasite_drag_nacelle(state,settings,geometry):
    """Computes the parasite drag due to the nacelle

    Assumptions:
        None 

    Source:
        None 

    Args: 
        state.conditions.freestream.mach_number      (numpy.ndarray): mach_number      [Unitless]
        state.conditions.freestream.temperature      (numpy.ndarray): temperature      [K]
        state.conditions.freestream.reynolds_number  (numpy.ndarray): Reynolds number  [Unitless]
        settings                                              (dict): analyses settings [-]
        geometry                                              (dict): aircraft geometry [-]

    Returns:
        None 
    """
     
    # Estimating nacelle drag 
    for network in  geometry.networks: 
        for propulsor in network.propulsors:  
            if 'nacelle' in propulsor: 
                nacelle_drag(state,settings,propulsor.nacelle)
    return     
# ---------------------------------------------------------------------------------------------------------------------- 
#  Nacelle Drag 
# ---------------------------------------------------------------------------------------------------------------------- 
def nacelle_drag(state,settings, nacelle):
    """helperr fuction to computes the parasite drag due to the nacelle

    Assumptions:
        None 

    Source:
        Stanford AA241 A/B Course Notes

    Args: 
        state.conditions.freestream.mach_number      (numpy.ndarray): mach_number      [Unitless]
        state.conditions.freestream.temperature      (numpy.ndarray): temperature      [K]
        state.conditions.freestream.reynolds_number  (numpy.ndarray): Reynolds number  [Unitless]
        settings                                              (dict): analyses settings [-] 

    Returns:
        None 
    """

    # unpack inputs
    conditions       = state.conditions
    freestream       = conditions.freestream
    Mach             = freestream.mach_number
    T                = freestream.temperature     
    Re               = freestream.reynolds_number
    low_mach_cutoff  = settings.supersonic.begin_drag_rise_mach_number
    high_mach_cutoff = settings.supersonic.end_drag_rise_mach_number 
    Sref             = nacelle.diameter**2 / 4 * np.pi
    Swet             = nacelle.areas.wetted
    
    # Reynolds number
    Re_prop = Re*nacelle.length
    
    # Skin friction coefficient
    cf_prop, k_comp, k_reyn = compressible_turbulent_flat_plate(Re_prop,Mach,T) 
    
    # Form factor according to Raymer equation
    form_factor  = 1 + 0.35 / ( nacelle.length/nacelle.diameter)   
         
    if np.all((Mach<=1.0) == True): 
        # subsonic condition 
        parasite_drag = form_factor * cf_prop * Swet / Sref 
    else:

        # supersonic condition 
        k_prop_sup = 1.
        
        trans_spline = Cubic_Spline_Blender(low_mach_cutoff,high_mach_cutoff)
        h00 = lambda M:trans_spline.compute(M)
        
        form_factor = form_factor*(h00(Mach)) + k_prop_sup*(1-h00(Mach))
             
        # find the final result    
        parasite_drag = form_factor * cf_prop * Swet / Sref        
    
    # store results
    results = Data(
        wetted_area               = Swet    , 
        reference_area            = Sref    , 
        total                     = parasite_drag ,
        skin_friction             = cf_prop ,
        compressibility_factor    = k_comp  ,
        reynolds_factor           = k_reyn  , 
        form_factor               = form_factor  ,
    )
    state.conditions.aerodynamics.coefficients.drag.parasite[nacelle.tag] = results    
    
    return



""" RCAIDE Package Setup for rotor wake methods that are directly specified by analyses.
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Wake.Perscribed_Vortex_Wake  import compute_fidelity_one_inflow_velocities 
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Wake.Perscribed_Vortex_Wake  import compute_wake_induced_velocity


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_fuselage_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# fuselage Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_fuselage_weight(vehicle, fuselage, settings):
    """ Calculate the weight of the fuselage of a transport aircraft based on the Raymer method

        Assumptions:
            No fuselage mounted landing gear
            1 cargo door

        Source:
            Aircraft Design: A ConceptualengthApproach (2nd edition)

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.mass_properties.max_takeoff: MTOW                             [kg]
                -.flight_envelope.ultimate_load: ultimate load factor (default: 3.75)
                -.wings['main_wing']: data dictionary with main wing properties
                    -.taper: wing taper ratio
                    -.sweeps.quarter_chord: quarter chord sweep                 [rad]
            fuselage - data dictionary with specific fuselage properties            [dimensionless]
                -.lenghts.total: totalengthlength                                   [m]
                -.width: fuselage width                                         [m]
                -.heights.maximum: maximum height of the fuselage               [m]

        Outputs:
            weight_fuselage - weight of the fuselage                                [kilograms]

        Properties Used:
            N/A
    """
    Klg         = settings.Raymer.fuselage_mounted_landing_gear_factor
    DG          = vehicle.mass_properties.max_takeoff / Units.lbs
    length      = fuselage.lengths.total/ Units.ft
    fuselage_w  = fuselage.width / Units.ft
    fuselage_h  = fuselage.heights.maximum / Units.ft
    
    Kdoor       = 1.06  # Assuming 1 cargo door
    D           = (fuselage_w + fuselage_h) / 2.
    Sf          = np.pi * (length/ D - 1.7) * D ** 2  # fuselage wetted area, ft**2
    wing        = vehicle.wings['main_wing']
    Kws         = 0.75 * (1 + 2 * wing.taper) / (1 + wing.taper) * (wing.spans.projected / Units.ft *
                                                            np.tan(wing.sweeps.quarter_chord)) / length

    weight_fuselage = 0.328 * Kdoor * Klg * (DG * vehicle.flight_envelope.ultimate_load) ** 0.5 * length** 0.25 * \
                 Sf ** 0.302 * (1 + Kws) ** 0.04 * (length/ D) ** 0.1
    return weight_fuselage * Units.lbs



# RCAIDE/Methods/Noise/Frequency_Domain_Buildup/Rotor/BPM_boundary_layer_properties.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports 
from RCAIDE.Framework.Core import  Data   

# Python package imports   
import numpy as np    

# ----------------------------------------------------------------------------------------------------------------------    
#  BPM_boundary_layer_properties
# ----------------------------------------------------------------------------------------------------------------------     
def BPM_boundary_layer_properties(R_c,c,alpha_star):
    '''This computes the boundary layer properties using the method outlined by the 
    Brooks, Pope and Marcolini (BPM) Model
    
    Assumptions:
        BPM models assumes a naca 0012 airfol  
        
    Source:   
        BPM Model:  Brooks, Thomas F., D. Stuart Pope, and Michael A.
        Marcolini. Airfoil self-noise and prediction. No. L-16528. 1989.
    
    Inputs:  
       a lpha_star - adjustd angle of attack    [deg] 
       c          - airfoil section chord       [m]
       R_c        - Reynolds number             [-]

    Outputs 
       boundary layer properties                [-]
       
    Properties Used:
        N/A   
    '''        
    # eqn 2
    delta_0_div_c_tripped = 10**(1.892- 0.9045*np.log10(R_c) + 0.0596*(np.log10(R_c))**2)
    
    # eqn 3  
    delta_star_0_div_c_tripped             = 0.0601*(R_c**-0.114) #R_c <= 0.3E6
    delta_star_0_div_c_tripped[R_c>0.3E6]  = 10**(3.411 -1.5397*np.log10(R_c[R_c>0.3E6]) + 0.1059*(np.log10(R_c[R_c>0.3E6]))**2) # R_c > 0.3E6 
    
    # eqn 4 
    theta_0_div_c_tripped             = 0.0723*(R_c**-0.1765) #R_c <= 0.3E6
    theta_0_div_c_tripped[R_c>0.3E6]  = 10**(0.5578 -0.7079*np.log10(R_c[R_c>0.3E6]) + 0.0404*(np.log10(R_c[R_c>0.3E6]))**2) # R_c > 0.3E6 
        
    # eqn 5
    delta_0_div_c_untripped       =  10**(1.6569 -0.9045*np.log10(R_c) + 0.0596*(np.log10(R_c))**2) 
    
    # eqn 6
    #delta_star_0_div_c_untripped  =  10**(3.0187 -1.5397*np.log10(R_c) + 0.1059*(np.log10(R_c))**2) 

    # eqn 7
    #theta_0_div_c_untripped       =  10**(0.2021 -0.7079*np.log10(R_c) + 0.0404*(np.log10(R_c))**2) 
    
    # boundary layer of pressure side for tripped and untripped 
    # eqn 8
    #delta_p_div_delta_0_untripped       =  10**(-0.04175*alpha_star + 0.00106*(alpha_star**2))
    #delta_p_div_delta_0_tripped         = delta_p_div_delta_0_untripped
    
    # eqn 9
    delta_star_p_div_delta_0_untripped  =  10**(-0.0432*alpha_star + 0.00113*(alpha_star**2))  
    delta_star_p_div_delta_0_tripped    = delta_star_p_div_delta_0_untripped
    
    # eqn 10
    #theta_p_div_delta_0_untripped       =  10**(-0.04408*alpha_star + 0.000873*(alpha_star**2))  
    #theta_p_div_delta_0_tripped         = theta_p_div_delta_0_untripped
    
    # boundary layer of suction side for tripped and untripped  
    # eqn 11
    delta_s_div_delta_0_tripped                   = 0.3468*(10**(0.1231*alpha_star))
    delta_s_div_delta_0_tripped[alpha_star<5]     = 10**(0.0311*alpha_star[alpha_star<5])
    delta_s_div_delta_0_tripped[alpha_star>12.5]  = 5.718*(10**(0.0258*alpha_star[alpha_star>12.5])) 
     
    # eqn 12  
    delta_star_s_div_delta_0_tripped                   = 0.381*(10**(0.1516*alpha_star))
    delta_star_s_div_delta_0_tripped[alpha_star<5]     = 10**(0.0679*alpha_star[alpha_star<5])
    delta_star_s_div_delta_0_tripped[alpha_star>12.5]  = 14.296*(10**(0.0258*alpha_star[alpha_star>12.5]))   
      
    # eqn 13 
    theta_s_div_delta_0_tripped                   = 0.6984*(10**(0.0869*alpha_star))
    theta_s_div_delta_0_tripped[alpha_star<5]     = 10**(0.0559*alpha_star[alpha_star<5])
    theta_s_div_delta_0_tripped[alpha_star>12.5]  = 4.0846*(10**(0.0258*alpha_star[alpha_star>12.5]))  

    # eqn 14  
    delta_s_div_delta_0_untripped                   = 0.0303*(10**(0.2336*alpha_star))
    delta_s_div_delta_0_untripped[alpha_star<7.5]   = 10**(0.03114*alpha_star[alpha_star<7.5])
    delta_s_div_delta_0_untripped[alpha_star>12.5]  = 12*(10**(0.0258*alpha_star[alpha_star>12.5]))  
     
    # eqn 15 
    delta_star_s_div_delta_0_untripped                   = 0.0162*(10**(0.3066*alpha_star))
    delta_star_s_div_delta_0_untripped[alpha_star<7.5]   = 10**(0.0672*alpha_star[alpha_star<7.5])
    delta_star_s_div_delta_0_untripped[alpha_star>12.5]  = 52.42*(10**(0.0258*alpha_star[alpha_star>12.5]))  
      
    # eqn 16     
    theta_s_div_delta_0_untripped                   = 0.0633*(10**(0.2157*alpha_star))
    theta_s_div_delta_0_untripped[alpha_star<7.5]   = 10**(0.0679*alpha_star[alpha_star<7.5])
    theta_s_div_delta_0_untripped[alpha_star>12.5]  = 14.977*(10**(0.0258*alpha_star[alpha_star>12.5]))   
         
    boundary_layer_data                              = Data() 
    
    # pressure side  
    boundary_layer_data.delta_star_p_untripped       = delta_star_p_div_delta_0_untripped*delta_0_div_c_untripped*c
    boundary_layer_data.delta_star_p_tripped         = delta_star_p_div_delta_0_tripped*delta_0_div_c_tripped*c
    
    # suction side  
    boundary_layer_data.delta_star_s_tripped         = delta_star_s_div_delta_0_tripped*delta_0_div_c_tripped*c
    boundary_layer_data.delta_star_s_untripped       = delta_star_s_div_delta_0_untripped*delta_0_div_c_untripped*c 
    return boundary_layer_data     


# RCAIDE/Library/Methods/Weights/Buildups/eVTOL/converge_physics_based_weight_buildup.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE 

# ----------------------------------------------------------------------------------------------------------------------
# converge_physics_based_weight_buildup
# ---------------------------------------------------------------------------------------------------------------------- 
def converge_physics_based_weight_buildup(base_vehicle,
                    print_iterations              = False,
                    miscelleneous_weight_factor   = 1.1,):
    '''Converges the maximum takeoff weight of an aircraft using the eVTOL 
    weight buildup routine.  
    
    Source:
    None
    
    Assumptions:
    None
    
    Inputs:
    vehicle                     RCAIDE Config Data Stucture
    print_iterations            Boolean Flag      
    miscelleneous_weight_factor          Factor capturing uncertainty in vehicle weight [Unitless]
    speed_of_sound:             Local Speed of Sound                           [m/s]
    max_tip_mach:               Allowable Tip Mach Number                      [Unitless]
    disk_area_factor:           Inverse of Disk Area Efficiency                [Unitless]
    max_thrust_to_weight_ratio: Allowable Thrust to Weight Ratio               [Unitless]
    safety_factor               Safety Factor in vehicle design                [Unitless]
    max_g_load                  Maximum g-forces load for certification        [UNitless]
    motor_efficiency:           Motor Efficiency                               [Unitless]
    
    Outputs:
    None
    
    Properties Used:
    N/A
    '''
    

    weight_analysis          = RCAIDE.Framework.Analyses.Weights.Weights_EVTOL()
    weight_analysis.vehicle  = base_vehicle
    weight_analysis.settings.miscelleneous_weight_factor =  miscelleneous_weight_factor
    breakdown                = weight_analysis.evaluate() 
    build_up_mass            = breakdown.total    
    diff                     = weight_analysis.vehicle.mass_properties.max_takeoff - build_up_mass
    iterations               = 0
    
    while(abs(diff)>1):
        weight_analysis.vehicle.mass_properties.max_takeoff = weight_analysis.vehicle.mass_properties.max_takeoff - diff 
        breakdown      = weight_analysis.evaluate()         
        build_up_mass  = breakdown.total    
        diff           = weight_analysis.vehicle.mass_properties.max_takeoff - build_up_mass 
        iterations     += 1
        if print_iterations:
            print(round(diff,3))
        if iterations == 100:
            print('Weight convergence failed!')
            return False 
    print('Converged MTOW = ' + str(round(weight_analysis.vehicle.mass_properties.max_takeoff)) + ' kg') 
    
    return weight_analysis.vehicle , breakdown
